import os.path
import sys
import textwrap

from PySide6.QtCore import Qt, QObject, QUrl, QDir, Signal, QPoint, QModelIndex
from PySide6.QtWidgets import (QApplication, QWidget, QGridLayout, QLabel, QLineEdit, QPushButton, QFileDialog,
                               QTextBrowser, QInputDialog, QMenu, QTreeView, QAbstractItemView, QTabWidget, QTableView,
                               QCheckBox)
from PySide6.QtGui import (QPixmap, QTextCursor, QStandardItem, QIcon, QDesktopServices, QAction,
                           QPainter, QColor, QImage, QStandardItemModel)
from functools import partial
from typing import Optional

from .gui_style import *
from .fmu_operations import *
from .assembly import Assembly, AssemblyNode
from .checker import checker_list
from .help import Help
from .version import __version__ as version


class DropZoneWidget(QLabel):
    WIDTH = 150
    HEIGHT = 150
    fmu = None
    last_directory = None
    clicked = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.set_image(None)
        self.setProperty("class", "dropped_fmu")
        self.setFixedSize(self.WIDTH, self.HEIGHT)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls():
            event.setDropAction(Qt.DropAction.CopyAction)
            try:
                file_path = event.mimeData().urls()[0].toLocalFile()
            except IndexError:
                print("Please select a regular file.")
                return
            self.set_fmu(file_path)
            event.accept()
        else:
            event.ignore()

    def mousePressEvent(self, event):
        if self.last_directory:
            default_directory = self.last_directory
        else:
            default_directory = os.path.expanduser('~')

        fmu_filename, _ = QFileDialog.getOpenFileName(parent=self, caption='Select FMU to Manipulate',
                                                      dir=default_directory, filter="FMU files (*.fmu)")
        if fmu_filename:
            self.set_fmu(fmu_filename)

    def set_image(self, filename=None):
        if not filename:
            filename = os.path.join(os.path.dirname(__file__), "resources", "drop_fmu.png")
        elif not os.path.isfile(filename):
            filename = os.path.join(os.path.dirname(__file__), "resources", "fmu.png")

        base_image = QImage(filename).scaled(self.WIDTH, self.HEIGHT, Qt.AspectRatioMode.IgnoreAspectRatio,
                                             Qt.TransformationMode.SmoothTransformation)
        mask_filename = os.path.join(os.path.dirname(__file__), "resources", "mask.png")
        mask_image = QImage(mask_filename).scaled(self.WIDTH, self.HEIGHT, Qt.AspectRatioMode.IgnoreAspectRatio,
                                                  Qt.TransformationMode.SmoothTransformation)
        rounded_image = QImage(self.WIDTH, self.HEIGHT, QImage.Format.Format_ARGB32)
        rounded_image.fill(QColor(0, 0, 0, 0))
        painter = QPainter()
        painter.begin(rounded_image)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.drawImage(QPoint(0, 0), base_image)
        painter.drawImage(QPoint(0, 0), mask_image)
        painter.end()
        pixmap = QPixmap.fromImage(rounded_image)

        self.setPixmap(pixmap)

    def set_fmu(self, filename: str):
        try:
            self.last_directory = os.path.dirname(filename)
            self.fmu = FMU(filename)
            self.set_image(os.path.join(self.fmu.tmp_directory, "model.png"))
        except Exception as e:
            print(f"ERROR: Cannot load this FMU: {e}")
            self.set_image(None)
            self.fmu = None
        self.clicked.emit()


class LogWidget(QTextBrowser):
    class XStream(QObject):
        _stdout = None
        _stderr = None
        messageWritten = Signal(str)

        def __init__(self):
            super().__init__()

        def flush(self):
            pass

        @staticmethod
        def fileno():
            return -1

        def write(self, msg):
            if not self.signalsBlocked():
                self.messageWritten.emit(msg)

        @staticmethod
        def stdout():
            if not LogWidget.XStream._stdout:
                LogWidget.XStream._stdout = LogWidget.XStream()
                sys.stdout = LogWidget.XStream._stdout
            return LogWidget.XStream._stdout

        @staticmethod
        def stderr():
            if not LogWidget.XStream._stderr:
                LogWidget.XStream._stderr = LogWidget.XStream()
                sys.stderr = LogWidget.XStream._stderr
            return LogWidget.XStream._stderr

    def __init__(self):
        super().__init__()
        self.setMinimumWidth(900)
        self.setMinimumHeight(500)
        self.setSearchPaths([os.path.join(os.path.dirname(__file__), "resources")])
        self.insertHtml('<center><img src="fmu_manipulation_toolbox.png"/></center><br/>')
        LogWidget.XStream.stdout().messageWritten.connect(self.insertPlainText)
        LogWidget.XStream.stderr().messageWritten.connect(self.insertPlainText)

    def loadResource(self, _, name):
        image_path = os.path.join(os.path.dirname(__file__), "resources", name.toString())
        return QPixmap(image_path)


class HelpWidget(QLabel):
    HELP_URL = "https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/README.md"

    def __init__(self):
        super().__init__()
        self.setProperty("class", "help")

        filename = os.path.join(os.path.dirname(__file__), "resources", "help.png")
        image = QPixmap(filename)
        self.setPixmap(image)
        self.setAlignment(Qt.AlignmentFlag.AlignRight)

    def mousePressEvent(self, event):
        QDesktopServices.openUrl(QUrl(self.HELP_URL))


class FilterWidget(QPushButton):
    def __init__(self, items: Optional[list[str]] = (), parent=None):
        super().__init__(parent)
        self.items_selected = set(items)
        self.nb_items = len(items)
        self.update_filter_text()
        if items:
            self.menu = QMenu()
            for item in items:
                action = QAction(item, self)
                action.setCheckable(True)
                action.setChecked(True)
                action.triggered.connect(partial(self.toggle_item, action))
                self.menu.addAction(action)
            self.setMenu(self.menu)

    def toggle_item(self, action: QAction):
        if not action.isChecked() and len(self.items_selected) == 1:
            action.setChecked(True)

        if action.isChecked():
            self.items_selected.add(action.text())
        else:
            self.items_selected.remove(action.text())

        self.update_filter_text()

    def update_filter_text(self):
        if len(self.items_selected) == self.nb_items:
            self.setText("All causalities")
        else:
            self.setText(", ".join(sorted(self.items_selected)))

    def get(self):
        if len(self.items_selected) == self.nb_items:
            return []
        else:
            return sorted(self.items_selected)


class AssemblyTreeWidget(QTreeView):
    class AssemblyTreeModel(QStandardItemModel):

        def __init__(self, assembly: Assembly, parent=None):
            super().__init__(parent)

            self.lastDroppedItems = []
            self.pendingRemoveRowsAfterDrop = False
            self.setHorizontalHeaderLabels(['col1'])
            self.dnd_target_node: Optional[AssemblyNode] = None

            self.icon_container = QIcon(os.path.join(os.path.dirname(__file__), 'resources', 'container.png'))
            self.icon_fmu = QIcon(os.path.join(os.path.dirname(__file__), 'resources', 'icon_fmu.png'))

            if assembly:
                self.add_node(assembly.root, self)

        def add_node(self, node: AssemblyNode, parent_item):
            # Add Container
            item = QStandardItem(self.icon_container, node.name)
            parent_item.appendRow(item)
            item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsDragEnabled |
                          Qt.ItemFlag.ItemIsDropEnabled)
            item.setData(node, role=Qt.ItemDataRole.UserRole + 1)
            item.setData("container", role=Qt.ItemDataRole.UserRole + 2)

            # Add FMU's
            children_name = node.children.keys()
            for fmu_name in node.fmu_names_list:
                if fmu_name not in children_name:
                    fmu_node = QStandardItem(self.icon_fmu, fmu_name)
                    fmu_node.setData(node, role=Qt.ItemDataRole.UserRole + 1)
                    fmu_node.setData("fmu", role=Qt.ItemDataRole.UserRole + 2)
                    fmu_node.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable |
                                      Qt.ItemFlag.ItemIsDragEnabled)
                    item.appendRow(fmu_node)

            # Add Sub-Containers
            for child in node.children.values():
                self.add_node(child, item)

        def insertRows(self, row, count, parent=QModelIndex()):
            self.dnd_target_node = parent.data(role=Qt.ItemDataRole.UserRole + 1)
            return super().insertRows(row, count, parent=parent)

        def removeRows(self, row, count, parent=QModelIndex()):
            if not self.dnd_target_node:
                print("NO DROP NODE!?")

            source_index = self.itemFromIndex(parent).child(row, 0).data(role=Qt.ItemDataRole.UserRole+1)
            print(f"{source_index} ==> {self.dnd_target_node.name}")

            self.dnd_target_node = None
            return super().removeRows(row, count, parent)

        def dropMimeData(self, data, action, row, column, parent: QModelIndex):
            if parent.column() < 0:  # Avoid to drop item as a sibling of the root.
                return False
            return super().dropMimeData(data, action, row, column, parent)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.model = self.AssemblyTreeModel(None)
        self.setModel(self.model)

        self.expandAll()
        self.setDropIndicatorShown(True)
        self.setDragDropOverwriteMode(False)
        self.setAcceptDrops(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self.setRootIsDecorated(True)
        self.setHeaderHidden(True)

    def load_container(self, filename):
        assembly = Assembly(filename)
        self.model = self.AssemblyTreeModel(assembly)
        self.setModel(self.model)

    def setTopIndex(self):
        topIndex = self.model.index(0, 0, self.rootIndex())
        print(topIndex.isValid(), topIndex.model())
        if topIndex.isValid():
            self.setCurrentIndex(topIndex)
            if self.layoutCheck:
                self.model.layoutChanged.disconnect(self.setTopIndex)
        else:
            if not self.layoutCheck:
                self.model.layoutChanged.connect(self.setTopIndex)
                self.layoutCheck = True


    def dragEnterEvent2(self, event):
        if event.mimeData().hasImage:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent2(self, event):
        if event.mimeData().hasImage:
            event.accept()
        else:
            event.ignore()

    def dropEvent2(self, event):
        if event.mimeData().hasImage:
            event.setDropAction(Qt.DropAction.CopyAction)
            try:
                file_path = event.mimeData().urls()[0].toLocalFile()
            except IndexError:
                print("Please select a regular file.")
                return
            print(f"DROP: {file_path}")
            event.accept()
        else:
            event.ignore()


class AssemblPropertiesWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.layout = QGridLayout()
        self.layout.setVerticalSpacing(4)
        self.layout.setHorizontalSpacing(4)
        self.layout.setContentsMargins(10, 10, 10, 10)
        self.setLayout(self.layout)

        mt_check = QCheckBox("Multi-Threaded", self)
        self.layout.addWidget(mt_check, 1, 0)

        profiling_check = QCheckBox("Profiling", self)
        self.layout.addWidget(profiling_check, 1, 1)

        auto_inputs_check = QCheckBox("Auto Inputs", self)
        self.layout.addWidget(auto_inputs_check, 0, 0)

        auto_outputs_check = QCheckBox("Auto Outputs", self)
        self.layout.addWidget(auto_outputs_check, 0, 1)

        auto_links_check = QCheckBox("Auto Links", self)
        self.layout.addWidget(auto_links_check, 0, 2)


class AssemblyTabWidget(QTabWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        table = AssemblPropertiesWidget(parent=self)
        self.addTab(table, "Properties")
        table = QTableView()
        self.addTab(table, "Links")
        table = QTableView()
        self.addTab(table, "Inputs")
        table = QTableView()
        self.addTab(table, "Outputs")
        table = QTableView()
        self.addTab(table, "Start values")

        self.tabBar().setDocumentMode(True)
        self.tabBar().setExpanding(True)


class WindowWithLayout(QWidget):
    def __init__(self, title: str):
        super().__init__(None)  # Do not set parent to have a separated window
        self.setWindowTitle(title)

        self.layout = QGridLayout()
        self.layout.setVerticalSpacing(4)
        self.layout.setHorizontalSpacing(4)
        self.layout.setContentsMargins(10, 10, 10, 10)
        self.setLayout(self.layout)


class MainWindow(WindowWithLayout):
    def __init__(self):
        super().__init__('FMU Manipulation Toolbox')

        self.dropped_fmu = DropZoneWidget()
        self.dropped_fmu.clicked.connect(self.update_fmu)

        self.layout.addWidget(self.dropped_fmu, 0, 0, 4, 1)

        self.fmu_title = QLabel()
        self.fmu_title.setProperty("class", "title")
        self.layout.addWidget(self.fmu_title, 0, 1, 1, 4)

        self.container_window = None
        #TODO: Container Window
        #container_button = QPushButton("Make a container")
        #container_button.setProperty("class", "quit")
        #container_button.clicked.connect(self.launch_container)
        #self.layout.addWidget(container_button, 4, 1, 1, 1)

        help_widget = HelpWidget()
        self.layout.addWidget(help_widget, 0, 5, 1, 1)

        # Operations
        self.help = Help()
        operations_list = [
            ("Save port names",    '-dump-csv',           'save',    OperationSaveNamesToCSV, {"prompt_file": "write"}),
            ("Rename ports from CSV", '-rename-from-csv',   'modify',  OperationRenameFromCSV, {"prompt_file": "read"}),
            ("Remove Toplevel",       '-remove-toplevel',    'modify',  OperationStripTopLevel),
            ("Remove Regexp",         '-remove-regexp',      'removal', OperationRemoveRegexp, {"prompt": "regexp"}),
            ("Keep only Regexp",      '-keep-only-regexp',   'removal', OperationKeepOnlyRegexp, {"prompt": "regexp"}),
            ("Save description.xml",  '-extract-descriptor', 'save',    None, {"func": self.save_descriptor}),
            ("Trim Until",            '-trim-until',         'modify',  OperationTrimUntil, {"prompt": "Prefix"}),
            ("Merge Toplevel",        '-merge-toplevel',     'modify',  OperationMergeTopLevel),
            ("Remove all",            '-remove-all',         'removal', OperationRemoveRegexp, {"arg": ".*"}),
            ("Remove sources",        '-remove-sources',     'removal', OperationRemoveSources),
            ("Add Win32 remoting",    '-add-remoting-win32', 'info',    OperationAddRemotingWin32),
            ("Add Win64 remoting",    '-add-remoting-win64', 'info',    OperationAddRemotingWin64),
            ("Add Win32 frontend",    '-add-frontend-win32', 'info',    OperationAddFrontendWin32),
            ("Add Win64 frontend",    '-add-frontend-win64', 'info',    OperationAddFrontendWin64),
            ("Check",                 '-check',              'info',    checker_list),
        ]

        width = 5
        line = 1
        for i, operation in enumerate(operations_list):
            col = i % width + 1
            line = int(i / width) + 1

            if len(operation) < 5:
                self.add_operation(operation[0], operation[1], operation[2], operation[3], line, col)
            else:
                self.add_operation(operation[0], operation[1], operation[2], operation[3], line, col, **operation[4])

        line += 1
        self.apply_filter_label = QLabel("Apply only on: ")
        self.layout.addWidget(self.apply_filter_label, line, 2, 1, 1, alignment=Qt.AlignmentFlag.AlignRight)
        self.set_tooltip(self.apply_filter_label, 'gui-apply-only')

        causality = ["parameter", "calculatedParameter", "input", "output", "local", "independent"]
        self.filter_list = FilterWidget(items=causality)
        self.layout.addWidget(self.filter_list, line, 3, 1, 3)
        self.filter_list.setProperty("class", "quit")

        # Text
        line += 1
        self.log_widget = LogWidget()
        self.layout.addWidget(self.log_widget, line, 0, 1, width + 1)

        # buttons
        line += 1

        reload_button = QPushButton('Reload')
        self.layout.addWidget(reload_button, 4, 0, 1, 1)
        reload_button.clicked.connect(self.reload_fmu)
        reload_button.setProperty("class", "quit")

        exit_button = QPushButton('Exit')
        self.layout.addWidget(exit_button, line, 0, 1, 2)
        exit_button.clicked.connect(self.close)
        exit_button.setProperty("class", "quit")

        save_log_button = QPushButton('Save log as')
        self.layout.addWidget(save_log_button, line, 2, 1, 2)
        save_log_button.clicked.connect(self.save_log)
        save_log_button.setProperty("class", "save")

        save_fmu_button = QPushButton('Save modified FMU as')
        self.layout.addWidget(save_fmu_button, line, 4, 1, 2)
        save_fmu_button.clicked.connect(self.save_fmu)
        save_fmu_button.setProperty("class", "save")
        self.set_tooltip(save_fmu_button, '-output')

        # show the window
        self.show()

    def closeEvent(self, event):
        if self.container_window:
            self.container_window.close()
            self.container_window = None
        event.accept()

    def launch_container(self):
        if not self.container_window:
            self.container_window = ContainerWindow(self)

    def closing_container(self):
        self.container_window = None

    def set_tooltip(self, widget, usage):
        widget.setToolTip("\n".join(textwrap.wrap(self.help.usage(usage))))

    def reload_fmu(self):
        if self.dropped_fmu.fmu:
            filename = self.dropped_fmu.fmu.fmu_filename
            self.dropped_fmu.fmu = None
            self.dropped_fmu.set_fmu(filename)

    def save_descriptor(self):
        if self.dropped_fmu.fmu:
            fmu = self.dropped_fmu.fmu
            filename, ok = QFileDialog.getSaveFileName(self, "Select a file",
                                                       os.path.dirname(fmu.fmu_filename),
                                                       "XML files (*.xml)")
            if ok and filename:
                fmu.save_descriptor(filename)

    def save_fmu(self):
        if self.dropped_fmu.fmu:
            fmu = self.dropped_fmu.fmu
            filename, ok = QFileDialog.getSaveFileName(self, "Select a file",
                                                       os.path.dirname(fmu.fmu_filename),
                                                       "FMU files (*.fmu)")
            if ok and filename:
                fmu.repack(filename)
                print(f"Modified version saved as {filename}.")

    def save_log(self):
        if self.dropped_fmu.fmu:
            default_dir = os.path.dirname(self.dropped_fmu.fmu.fmu_filename)
        else:
            default_dir = None
        filename, ok = QFileDialog.getSaveFileName(self, "Select a file",
                                                   default_dir,
                                                   "TXT files (*.txt)")
        if ok and filename:
            try:
                with open(filename, "wt") as file:
                    file.write(str(self.log_widget.toPlainText()))
            except Exception as e:
                print(f"ERROR: {e}")

    def add_operation(self, name, usage, severity, operation, x, y, prompt=None, prompt_file=None, arg=None,
                      func=None):
        if prompt:
            def operation_handler():
                local_arg = self.prompt_string(prompt)
                if local_arg:
                    self.apply_operation(operation(local_arg))
        elif prompt_file:
            def operation_handler():
                local_arg = self.prompt_file(prompt_file)
                if local_arg:
                    self.apply_operation(operation(local_arg))
        elif arg:
            def operation_handler():
                self.apply_operation(operation(arg))
        else:
            def operation_handler():
                # Checker can be a list of operations!
                if isinstance(operation, list):
                    for op in operation:
                        self.apply_operation(op())
                else:
                    self.apply_operation(operation())

        button = QPushButton(name)
        self.set_tooltip(button, usage)
        button.setProperty("class", severity)
        if func:
            button.clicked.connect(func)
        else:
            button.clicked.connect(operation_handler)
        self.layout.addWidget(button, x, y)

    def prompt_string(self, message):
        text, ok = QInputDialog().getText(self, "Enter value", f"{message}:", QLineEdit.EchoMode.Normal, "")

        if ok and text:
            return text
        else:
            return None

    def prompt_file(self, access):
        if self.dropped_fmu.fmu:
            default_dir = os.path.dirname(self.dropped_fmu.fmu.fmu_filename)

            if access == 'read':
                filename, ok = QFileDialog.getOpenFileName(self, "Select a file",
                                                           default_dir, "CSV files (*.csv)")
            else:
                filename, ok = QFileDialog.getSaveFileName(self, "Select a file",
                                                           default_dir, "CSV files (*.csv)")

            if ok and filename:
                return filename
        return None

    def update_fmu(self):
        if self.dropped_fmu.fmu:
            self.fmu_title.setText(os.path.basename(self.dropped_fmu.fmu.fmu_filename))
            self.log_widget.clear()
            self.apply_operation(OperationSummary())
        else:
            self.fmu_title.setText('')

    def apply_operation(self, operation):
        if self.dropped_fmu.fmu:
            self.log_widget.moveCursor(QTextCursor.MoveOperation.End)
            fmu_filename = os.path.basename(self.dropped_fmu.fmu.fmu_filename)
            print('-' * 100)
            self.log_widget.insertHtml(f"<strong>{fmu_filename}: {operation}</strong><br>")

            apply_on = self.filter_list.get()
            if apply_on:
                self.log_widget.insertHtml(f"<i>Applied only for ports with  causality = " +
                                           ", ".join(apply_on) + "</i><br>")
            print('-' * 100)
            try:
                self.dropped_fmu.fmu.apply_operation(operation, apply_on=apply_on)
            except Exception as e:
                print(f"ERROR: {e}")

            scroll_bar = self.log_widget.verticalScrollBar()
            scroll_bar.setValue(scroll_bar.maximum())


class ContainerWindow(WindowWithLayout):
    def __init__(self, parent: MainWindow):
        super().__init__('FMU Manipulation Toolbox - Container')
        self.main_window = parent
        self.last_directory = None

        # ROW 0
        load_button = QPushButton("Load Description")
        load_button.clicked.connect(self.load_container)
        load_button.setProperty("class", "quit")
        self.layout.addWidget(load_button, 0, 0)

        self.container_label = QLabel()
        self.container_label.setProperty("class", "title")
        self.container_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.layout.addWidget(self.container_label, 0, 1, 1, 2)

        # ROW 1
        add_fmu_button = QPushButton("Add FMU")
        add_fmu_button.setProperty("class", "modify")
        add_fmu_button.setDisabled(True)
        self.layout.addWidget(add_fmu_button, 1, 1)

        add_sub_button = QPushButton("Add SubContainer")
        add_sub_button.setProperty("class", "modify")
        add_sub_button.setDisabled(True)
        self.layout.addWidget(add_sub_button, 1, 2)

        self.assembly_tree = AssemblyTreeWidget(parent=self)
        self.assembly_tree.setMinimumHeight(600)
        self.assembly_tree.setMinimumWidth(200)
        self.layout.addWidget(self.assembly_tree, 1, 0, 3, 1)

        # ROW 2
        del_fmu_button = QPushButton("Remove FMU")
        del_fmu_button.setProperty("class", "removal")
        del_fmu_button.setDisabled(True)
        self.layout.addWidget(del_fmu_button, 2, 1)

        del_sub_button = QPushButton("Remove SubContainer")
        del_sub_button.setProperty("class", "removal")
        del_sub_button.setDisabled(True)
        self.layout.addWidget(del_sub_button, 2, 2)

        # ROW 3
        self.assembly_tab = AssemblyTabWidget(parent=self)
        self.assembly_tab.setMinimumWidth(600)
        self.layout.addWidget(self.assembly_tab, 3, 1, 1, 2)

        # ROW 4
        close_button = QPushButton("Close")
        close_button.setProperty("class", "quit")
        close_button.clicked.connect(self.close)
        self.layout.addWidget(close_button, 4, 0)

        save_button = QPushButton("Save Container")
        save_button.setProperty("class", "save")
        self.layout.addWidget(save_button, 4, 2)

        self.assembly_tree.selectionModel().currentChanged.connect(self.item_selected)
        topIndex = self.assembly_tree.model.index(0, 0, self.assembly_tree.rootIndex())
        self.assembly_tree.setCurrentIndex(topIndex)

        self.show()

    def closeEvent(self, event):
        if self.main_window:
            self.main_window.closing_container()
        event.accept()

    def item_selected(self, current: QModelIndex, previous: QModelIndex):
        if current.isValid():
            node = current.data(role=Qt.ItemDataRole.UserRole + 1)
            node_type = current.data(role=Qt.ItemDataRole.UserRole + 2)
            self.container_label.setText(f"{node.name} ({node_type})")
        else:
            self.container_label.setText("")

    def load_container(self):
        if self.last_directory:
            default_directory = self.last_directory
        else:
            default_directory = os.path.expanduser('~')

        filename, _ = QFileDialog.getOpenFileName(parent=self, caption='Select FMU to Manipulate',
                                                  dir=default_directory,
                                                  filter="JSON files (*.json);;SSP files (*.ssp)")
        if filename:
            try:
                self.last_directory = os.path.dirname(filename)
                self.assembly_tree.load_container(filename)
            except Exception as e:
                print(e)


class Application(QApplication):
    """
Analyse and modify your FMUs.

Note: modifying the modelDescription.xml can damage your FMU !
Communicating with the FMU-developer and adapting the way the FMU is generated, is preferable when possible.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setHighDpiScaleFactorRoundingPolicy(Qt.HighDpiScaleFactorRoundingPolicy.RoundPreferFloor)

        QDir.addSearchPath('images', os.path.join(os.path.dirname(__file__), "resources"))
        self.setStyleSheet(gui_style_dark)

        if os.name == 'nt':
            import ctypes
            self.setWindowIcon(QIcon(os.path.join(os.path.dirname(__file__), 'resources', 'icon-round.png')))

            # https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105

            application_id = 'FMU_Manipulation_Toolbox'  # arbitrary string
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(application_id)
        else:
            self.setWindowIcon(QIcon(os.path.join(os.path.dirname(__file__), 'resources', 'icon.png')))

        self.window = MainWindow()


def main():
    application = Application(sys.argv)

    print(" " * 80, f"Version {version}")
    print(application.__doc__)

    sys.exit(application.exec())


if __name__ == "__main__":
    main()



import unittest
import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent))
from fmu_manipulation_toolbox.fmu_operations import *
from fmu_manipulation_toolbox.fmu_container import *
from fmu_manipulation_toolbox.assembly import *


class FMUManipulationToolboxTestSuite(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fmu_filename = "operations/bouncing_ball.fmu"

    def assert_identical_files(self, filename1, filename2):
        with open(filename1, mode="rt", newline=None) as a, open(filename2, mode="rt", newline=None) as b:
            self.assertTrue(all(lineA == lineB for lineA, lineB in zip(a, b)))

    def assert_names_match_ref(self, fmu_filename):
        fmu = FMU(fmu_filename)
        csv_filename = Path(fmu_filename).with_suffix(".csv")
        ref_filename = csv_filename.with_stem("REF-"+csv_filename.stem)
        operation = OperationSaveNamesToCSV(csv_filename)
        fmu.apply_operation(operation)
        self.assert_identical_files(ref_filename, csv_filename)

    def assert_operation_match_ref(self, fmu_filename, operation):
        fmu = FMU(self.fmu_filename)
        fmu.apply_operation(operation)
        fmu.repack(fmu_filename)
        self.assert_names_match_ref(fmu_filename)

    def test_strip_top_level(self):
        self.assert_operation_match_ref("operations/bouncing_ball-no-tl.fmu", OperationStripTopLevel())

    def test_save_names_to_CSV(self):
        self.assert_names_match_ref("operations/bouncing_ball.fmu")

    def test_rename_from_CSV(self):
        self.assert_operation_match_ref("operations/bouncing_ball-renamed.fmu",
                                        OperationRenameFromCSV("operations/bouncing_ball-modified.csv"))

    @unittest.skipUnless(sys.platform.startswith("win"), "Supported only on Windows")
    def test_add_remoting_win32(self):
        fmu = FMU(self.fmu_filename)
        operation = OperationAddRemotingWin32()
        fmu.apply_operation(operation)
        fmu.repack("bouncing_ball-win32.fmu")

    def test_remove_regexp(self):
        self.assert_operation_match_ref("operations/bouncing_ball-removed.fmu",
                                        OperationRemoveRegexp("e"))

    def test_keep_only_regexp(self):
        self.assert_operation_match_ref("operations/bouncing_ball-keeponly.fmu",
                                        OperationKeepOnlyRegexp("e"))

    def test_container_bouncing_ball(self):
        assembly = Assembly("bouncing.csv", fmu_directory=Path("containers/bouncing_ball"), mt=True, debug=True)
        assembly.write_json("bouncing.json")
        assembly.make_fmu()
        self.assert_identical_files("containers/bouncing_ball/REF-container.txt",
                                    "containers/bouncing_ball/bouncing/resources/container.txt")
        self.assert_identical_files("containers/bouncing_ball/REF-bouncing.json",
                                    "containers/bouncing_ball/bouncing.json")

    def test_container_ssp(self):
        assembly = Assembly("bouncing.ssp", fmu_directory=Path("containers/ssp"))
        assembly.make_fmu(dump_json=True)
        self.assert_identical_files("containers/ssp/REF-bouncing-dump.json",
                                    "containers/ssp/bouncing-dump.json")

    def test_container_json_flat(self):
        assembly = Assembly("flat.json", fmu_directory=Path("containers/arch"))
        assembly.make_fmu(dump_json=True)
        self.assert_identical_files("containers/arch/REF-flat-dump.json",
                                    "containers/arch/flat-dump.json")

    def test_container_json_hierarchical(self):
        assembly = Assembly("hierarchical.json", fmu_directory=Path("containers/arch"))
        assembly.make_fmu(dump_json=True)
        self.assert_identical_files("containers/arch/REF-hierarchical-dump.json",
                                    "containers/arch/hierarchical-dump.json")

    def test_container_json_reversed(self):
        assembly = Assembly("reversed.json", fmu_directory=Path("containers/arch"))
        assembly.make_fmu(dump_json=True)
        self.assert_identical_files("containers/arch/REF-reversed-dump.json",
                                    "containers/arch/reversed-dump.json")

    def test_container_move(self):
        #bb = Assembly("bouncing.csv", fmu_directory=Path("containers/bouncing_ball"))
        #links = bb.root.get_fmu_connections("bb_position.fmu")
        #print("Links: ", links)
        #bb.write_json("bouncing.json")
        assembly = Assembly("nested.json", fmu_directory=Path("containers/arch"))
        fmu_name = "fmu1b.fmu"
        links_fmu1b = assembly.root.children["level1.fmu"].get_fmu_connections("fmu1b.fmu")

        print("RESULTS:")
        for link in links_fmu1b:
            print(f"{link}")

        links_fmu0a = assembly.root.get_fmu_connections("fmu0a.fmu")
        print("RESULTS:")
        for link in links_fmu0a:
            print(f"{link}")


if __name__ == '__main__':
    unittest.main()



try:
    from fmu_manipulation_toolbox.__version__ import __doc__ as __version__
except ModuleNotFoundError:
    __version__ = "0.0.dev0"

__author__ = "Nicolas.LAURENT@Renault.com"
__copyright__ = "Copyright 2023-2024, Renault SAS"
__license__ = """This code is released under the 2-Clause BSD license. 
See https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/LICENSE.txt"""



import csv
import html
import os
import re
import shutil
import tempfile
import xml.parsers.expat
import zipfile
import hashlib
from pathlib import Path


class FMU:
    """Unpack and Repack facilities for FMU package. Once unpacked, we can process Operation on
    modelDescription.xml file."""
    def __init__(self, fmu_filename):
        self.fmu_filename = fmu_filename
        self.tmp_directory = tempfile.mkdtemp()

        try:
            with zipfile.ZipFile(self.fmu_filename) as zin:
                zin.extractall(self.tmp_directory)
        except FileNotFoundError:
            raise FMUException(f"'{fmu_filename}' does not exist")
        self.descriptor_filename = os.path.join(self.tmp_directory, "modelDescription.xml")
        if not os.path.isfile(self.descriptor_filename):
            raise FMUException(f"'{fmu_filename}' is not valid: {self.descriptor_filename} not found")

    def __del__(self):
        shutil.rmtree(self.tmp_directory)

    def save_descriptor(self, filename):
        shutil.copyfile(os.path.join(self.tmp_directory, "modelDescription.xml"), filename)

    def repack(self, filename):
        with zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED) as zout:
            for root, dirs, files in os.walk(self.tmp_directory):
                for file in files:
                    zout.write(os.path.join(root, file),
                               os.path.relpath(os.path.join(root, file), self.tmp_directory))
        # TODO: Add check on output file

    def apply_operation(self, operation, apply_on=None):
        manipulation = Manipulation(operation, self)
        manipulation.manipulate(self.descriptor_filename, apply_on)


class FMUException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __repr__(self):
        return self.reason


class Manipulation:
    """Parse modelDescription.xml file and create a modified version"""
    def __init__(self, operation, fmu):
        self.output_filename = tempfile.mktemp()
        self.out = None
        self.operation = operation
        self.parser = xml.parsers.expat.ParserCreate()
        self.parser.StartElementHandler = self.start_element
        self.parser.EndElementHandler = self.end_element
        self.parser.CharacterDataHandler = self.char_data
        self.skip_until = None
        self.operation.set_fmu(fmu)

        self.current_port = 0
        self.port_translation = []
        self.port_name = []
        self.apply_on = None

    @staticmethod
    def escape(value):
        if isinstance(value, str):
            return html.escape(html.unescape(value))
        else:
            return value

    def start_element(self, name, attrs):
        if self.skip_until:
            return
        try:
            if name == 'ScalarVariable':
                causality = OperationAbstract.scalar_get_causality(attrs)
                if not self.apply_on or causality in self.apply_on:
                    if self.operation.scalar_attrs(attrs):
                        self.remove_port(attrs['name'])
                    else:
                        self.keep_port(attrs['name'])
                else:  # Keep ScalarVariable as it is.
                    self.keep_port(attrs['name'])
            elif name == 'CoSimulation':
                self.operation.cosimulation_attrs(attrs)
            elif name == 'DefaultExperiment':
                self.operation.experiment_attrs(attrs)
            elif name == 'fmiModelDescription':
                self.operation.fmi_attrs(attrs)
            elif name == 'Unknown':
                self.unknown_attrs(attrs)
            elif name in ('Real', 'Integer', 'String', 'Boolean', 'Enumeration'):
                self.operation.scalar_type(name, attrs)

        except ManipulationSkipTag:
            self.skip_until = name
            return

        if attrs:
            attrs_list = [f'{key}="{self.escape(value)}"' for (key, value) in attrs.items()]
            print(f"<{name}", " ".join(attrs_list), ">", end='', file=self.out)
        else:
            print(f"<{name}>", end='', file=self.out)

    def end_element(self, name):
        if self.skip_until:
            if self.skip_until == name:
                self.skip_until = None
            return
        else:
            print(f"</{name}>", end='', file=self.out)

    def char_data(self, data):
        if not self.skip_until:
            print(data, end='', file=self.out)

    def remove_port(self, name):
        self.port_name.append(name)
        self.port_translation.append(None)
        raise ManipulationSkipTag

    def keep_port(self, name):
        self.port_name.append(name)
        self.current_port += 1
        self.port_translation.append(self.current_port)

    def unknown_attrs(self, attrs):
        index = int(attrs['index']) - 1
        new_index = self.port_translation[index]
        if new_index:
            attrs['index'] = self.port_translation[int(attrs['index']) - 1]
        else:
            print(f"WARNING: Removed port '{self.port_name[index]}' is involved in dependencies tree.")
            raise ManipulationSkipTag

    def manipulate(self, descriptor_filename, apply_on=None):
        self.apply_on = apply_on
        with open(self.output_filename, "w", encoding="utf-8") as self.out, open(descriptor_filename, "rb") as file:
            self.parser.ParseFile(file)
        self.operation.closure()
        os.replace(self.output_filename, descriptor_filename)


class ManipulationSkipTag(Exception):
    """Exception: We need to skip every thing until matching closing tag"""


class OperationAbstract:
    """This class hold hooks called during parsing"""
    fmu: FMU = None

    def set_fmu(self, fmu):
        self.fmu = fmu

    def fmi_attrs(self, attrs):
        pass

    def scalar_attrs(self, attrs) -> int:
        """ return 0 to keep port, otherwise remove it"""
        return 0

    def cosimulation_attrs(self, attrs):
        pass

    def experiment_attrs(self, attrs):
        pass

    def scalar_type(self, type_name, attrs):
        pass

    def closure(self):
        pass

    @staticmethod
    def scalar_get_causality(attrs) -> str:
        try:
            causality = attrs['causality']
        except KeyError:
            causality = 'local'  # Default value according to FMI Specifications.

        return causality


class OperationSaveNamesToCSV(OperationAbstract):
    def __repr__(self):
        return f"Dump names into '{self.output_filename}'"

    def __init__(self, filename):
        self.output_filename = filename
        self.csvfile = open(filename, 'w', newline='')
        self.writer = csv.writer(self.csvfile, delimiter=';', quotechar="'", quoting=csv.QUOTE_MINIMAL)
        self.writer.writerow(['name', 'newName', 'valueReference', 'causality', 'variability', 'scalarType',
                              'startValue'])
        self.name = None
        self.vr = None
        self.variability = None
        self.causality = None

    def reset(self):
        self.name = None
        self.vr = None
        self.variability = None
        self.causality = None

    def closure(self):
        self.csvfile.close()

    def scalar_attrs(self, attrs):
        self.name = attrs['name']
        self.vr = attrs['valueReference']
        self.causality = self.scalar_get_causality(attrs)

        try:
            self.variability = attrs['variability']
        except KeyError:
            self.variability = 'continuous'   # Default value according to FMI Specifications.

        return 0

    def scalar_type(self, type_name, attrs):
        if "start" in attrs:
            start = attrs["start"]
        else:
            start = ""
        self.writer.writerow([self.name, self.name, self.vr, self.causality, self.variability, type_name, start])
        self.reset()


class OperationStripTopLevel(OperationAbstract):
    def __repr__(self):
        return "Remove Top Level Bus"

    def scalar_attrs(self, attrs):
        new_name = attrs['name'].split('.', 1)[-1]
        attrs['name'] = new_name
        return 0


class OperationMergeTopLevel(OperationAbstract):
    def __repr__(self):
        return "Merge Top Level Bus with signal names"

    def scalar_attrs(self, attrs):
        old = attrs['name']
        attrs['name'] = old.replace('.', '_', 1)
        return 0


class OperationRenameFromCSV(OperationAbstract):
    def __repr__(self):
        return f"Rename according to '{self.csv_filename}'"

    def __init__(self, csv_filename):
        self.csv_filename = csv_filename
        self.translations = {}
        self.current_port = 0
        self.port_translation = []
        try:
            with open(csv_filename, newline='') as csvfile:
                reader = csv.reader(csvfile, delimiter=';', quotechar="'")
                for row in reader:
                    self.translations[row[0]] = row[1]
        except FileNotFoundError:
            raise OperationException(f"file '{csv_filename}' is not found")
        except KeyError:
            raise OperationException(f"file '{csv_filename}' should contain two columns")

    def scalar_attrs(self, attrs):
        name = attrs['name']
        try:
            new_name = self.translations[attrs['name']]
        except KeyError:
            new_name = name  # if port is not in CSV file, keep old name

        if new_name:
            attrs['name'] = new_name
            return 0
        else:
            # we want to delete this name!
            return 1


class OperationAddRemotingWinAbstract(OperationAbstract):
    bitness_from = None
    bitness_to = None

    def __repr__(self):
        return f"Add '{self.bitness_to}' remoting on '{self.bitness_from}' FMU"

    def cosimulation_attrs(self, attrs):
        fmu_bin = {
            "win32":  os.path.join(self.fmu.tmp_directory, "binaries", f"win32"),
            "win64": os.path.join(self.fmu.tmp_directory, "binaries", f"win64"),
        }

        if not os.path.isdir(fmu_bin[self.bitness_from]):
            raise OperationException(f"{self.bitness_from} interface does not exist")

        if os.path.isdir(fmu_bin[self.bitness_to]):
            print(f"INFO: {self.bitness_to} already exists. Add front-end.")
            shutil.move(os.path.join(fmu_bin[self.bitness_to], attrs['modelIdentifier'] + ".dll"),
                        os.path.join(fmu_bin[self.bitness_to], attrs['modelIdentifier'] + "-remoted.dll"))
        else:
            os.mkdir(fmu_bin[self.bitness_to])

        from_path = Path(__file__).parent / "resources" / self.bitness_to
        shutil.copyfile(from_path / "client_sm.dll",
                        Path(fmu_bin[self.bitness_to]) / Path(attrs['modelIdentifier']).with_suffix(".dll"))

        shutil.copyfile(from_path / "server_sm.exe",
                        Path(fmu_bin[self.bitness_from]) / "server_sm.exe")

        shutil.copyfile(Path(__file__).parent / "resources" / "license.txt",
                        Path(fmu_bin[self.bitness_to]) / "license.txt")


class OperationAddRemotingWin64(OperationAddRemotingWinAbstract):
    bitness_from = "win32"
    bitness_to = "win64"


class OperationAddFrontendWin32(OperationAddRemotingWinAbstract):
    bitness_from = "win32"
    bitness_to = "win32"


class OperationAddFrontendWin64(OperationAddRemotingWinAbstract):
    bitness_from = "win64"
    bitness_to = "win64"


class OperationAddRemotingWin32(OperationAddRemotingWinAbstract):
    bitness_from = "win64"
    bitness_to = "win32"


class OperationRemoveRegexp(OperationAbstract):
    def __repr__(self):
        return f"Remove ports matching '{self.regex_string}'"

    def __init__(self, regex_string):
        self.regex_string = regex_string
        self.regex = re.compile(regex_string)
        self.current_port = 0
        self.port_translation = []

    def scalar_attrs(self, attrs):
        name = attrs['name']
        if self.regex.match(name):
            return 1  # Remove port
        else:
            return 0


class OperationKeepOnlyRegexp(OperationAbstract):
    def __repr__(self):
        return f"Keep only ports matching '{self.regex_string}'"

    def __init__(self, regex_string):
        self.regex_string = regex_string
        self.regex = re.compile(regex_string)

    def scalar_attrs(self, attrs):
        name = attrs['name']
        if self.regex.match(name):
            return 0
        else:
            return 1  # Remove port


class OperationSummary(OperationAbstract):
    def __init__(self):
        self.nb_port_per_causality = {}

    def __repr__(self):
        return f"FMU Summary"

    def fmi_attrs(self, attrs):
        print(f"| fmu filename = {self.fmu.fmu_filename}")
        print(f"| temporary directory = {self.fmu.tmp_directory}")
        hash_md5 = hashlib.md5()
        with open(self.fmu.fmu_filename, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        digest = hash_md5.hexdigest()
        print(f"| MD5Sum = {digest}")

        print(f"|\n| FMI properties: ")
        for (k, v) in attrs.items():
            print(f"|  - {k} = {v}")
        print(f"|")

    def cosimulation_attrs(self, attrs):
        print("| Co-Simulation capabilities: ")
        for (k, v) in attrs.items():
            print(f"|  - {k} = {v}")
        print(f"|")

    def experiment_attrs(self, attrs):
        print("| Default Experiment values: ")
        for (k, v) in attrs.items():
            print(f"|  - {k} = {v}")
        print(f"|")

    def scalar_attrs(self, attrs) -> int:
        causality = self.scalar_get_causality(attrs)

        try:
            self.nb_port_per_causality[causality] += 1
        except KeyError:
            self.nb_port_per_causality[causality] = 1

        return 0

    def closure(self):
        print("| Supported platforms: ")
        try:
            for platform in os.listdir(os.path.join(self.fmu.tmp_directory, "binaries")):
                print(f"|  - {platform}")
        except FileNotFoundError:
            pass  # no binaries

        if os.path.isdir(os.path.join(self.fmu.tmp_directory, "sources")):
            print(f"|  - RT (sources available)")

        resource_dir = os.path.join(self.fmu.tmp_directory, "resources")
        if os.path.isdir(resource_dir):
            print("|\n| Embedded resources:")
            for resource in os.listdir(resource_dir):
                print(f"|  - {resource}")

        extra_dir = os.path.join(self.fmu.tmp_directory, "extra")
        if os.path.isdir(extra_dir):
            print("|\n| Additional (meta-)data:")
            for extra in os.listdir(extra_dir):
                print(f"|  - {extra}")

        print("|\n| Number of signals")
        for causality, nb_ports in self.nb_port_per_causality.items():
            print(f"|  {causality} : {nb_ports}")

        print("|\n| [End of report]")


class OperationRemoveSources(OperationAbstract):
    def __repr__(self):
        return f"Remove sources"

    def cosimulation_attrs(self, attrs):
        try:
            shutil.rmtree(os.path.join(self.fmu.tmp_directory, "sources"))
        except FileNotFoundError:
            print("This FMU does not embed sources.")


class OperationTrimUntil(OperationAbstract):
    def __init__(self, separator):
        self.separator = separator

    def __repr__(self):
        return f"Trim names until (and including) '{self.separator}'"

    def scalar_attrs(self, attrs) -> int:
        name = attrs['name']
        try:
            attrs['name'] = name[name.index(self.separator)+len(self.separator):-1]
        except KeyError:
            pass  # no separator

        return 0


class OperationException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __repr__(self):
        return self.reason



# if os.name == 'nt':
#    font = QFont('Consolas')
#    font.setPointSize(11)
# else:
#    font = QFont('Courier New')
#    font.setPointSize(12)
# self.setFont(font)


gui_style_dark = """
QWidget {
    font: 10pt "Verdana";
    background: #4b4e51;
    color: #b5bab9;
}
QPushButton, QComboBox {
    min-height: 30px;
    padding: 1px 1px 0.2em 0.2em;
    border: 1px solid #282830;
    border-radius: 5px;
    color: #dddddd;
}
QPushButton:pressed {
    border: 2px solid #282830;
}
QPushButton.info {
    background-color: #4e6749;
}
QPushButton.info:hover {
    background-color: #5f7850;
}
QPushButton.modify {
    background-color: #98763f;
}
QPushButton.modify:hover {
    background-color: #a9874f;
}
QPushButton.removal {
    background-color: #692e2e;
}
QPushButton.removal:hover {
    background-color: #7a3f3f;
}
QPushButton.save {
    background-color: #564967;
}
QPushButton.save:hover {
    background-color: #675a78;
}
QPushButton.quit {
    background-color: #4571a4;
}
QPushButton.quit:hover {
    background-color: #5682b5;
}
QPushButton::disabled {
    background-color: gray;
}
QToolTip {
    color: black
}
QLabel.dropped_fmu {
    background-color: #b5bab9
}
QLabel.title {
    font: 14pt bold "Verdana";
}
QLabel.dropped_fmu:hover {
    background-color: #c6cbca
}
QTextBrowser, QTreeView {
    font: 11pt "Consolas";
    background-color: #282830;
    color: #b5bab9;
}
QMenu::item {
    padding: 2px 250px 2px 20px;
    border: 1px solid transparent;
}
QMenu::item::indicator, QCheckBox::item::indicator {
    width: 32px;
    height: 32px;
}
QMenu::indicator:checked, QCheckBox::indicator:checked {
    image: url(images:checkbox-checked.png);
}
QMenu::indicator:checked:hover, QCheckBox::indicator:checked:hover {
    image: url(images:checkbox-checked-hover.png);
}
QMenu::indicator:checked:disabled, QCheckBox::indicator:checked:disabled {
    image: url(images:checkbox-checked-disabled.png);
}
QMenu::indicator:unchecked, QCheckBox::indicator:unchecked {
    image: url(images:checkbox-unchecked.png);
}
QMenu::indicator:unchecked:hover, QCheckBox::indicator:unchecked:hover {
    image: url(images:checkbox-unchecked-hover.png);
}
QMenu::indicator:unchecked:disabled, QCheckBox::indicator:unchecked:disabled {
    image: url(images:checkbox-unchecked-disabled.png);
}
QCheckBox::item {
    padding: 2px 250px 2px 20px;
    border: 1px solid transparent;
}
QTabBar::tab {
    min-height: 30px;
    padding: 1px 1px 0.2em 0.2em;
    color: #dddddd;
    margin: 2px;
    margin-bottom: 0px;
    border: 1px solid #282830;
    border-top-left-radius: 5px;
    border-top-right-radius: 5px;
} 
QTabBar::tab:selected, QTabBar::tab:hover {
    background-color: #5f7850;
    margin-bottom:-1px;
}
QTabBar {
    border-bottom: 1px solid #282830;
}
QTabBar::tab:top:last, QTabBar::tab:bottom:last {
    margin-right: 0;
}
QTabBar::tab:top:first, QTabBar::tab:bottom:first {
    margin-left: 0;
}
"""



import argparse
import logging
import sys
from colorama import Fore, Style, init

from .fmu_operations import *
from .fmu_container import FMUContainerError
from .assembly import Assembly, AssemblyError
from .checker import checker_list
from .version import __version__ as version
from .help import Help


def setup_logger():
    class CustomFormatter(logging.Formatter):
        def format(self, record):
            log_format = "%(levelname)-8s | %(message)s"
            format_per_level = {
                logging.DEBUG: str(Fore.BLUE) + log_format,
                logging.INFO: str(Fore.CYAN) + log_format,
                logging.WARNING: str(Fore.YELLOW) + log_format,
                logging.ERROR: str(Fore.RED) + log_format,
                logging.CRITICAL: str(Fore.RED + Style.BRIGHT) + log_format,
            }
            formatter = logging.Formatter(format_per_level[record.levelno])
            return formatter.format(record)
    init()
    logger = logging.getLogger("fmu_manipulation_toolbox")
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(CustomFormatter())
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    return logger


def make_wide(formatter, w=120, h=36):
    """Return a wider HelpFormatter, if possible."""
    try:
        # https://stackoverflow.com/a/5464440
        # beware: "Only the name of this class is considered a public API."
        kwargs = {'width': w, 'max_help_position': h}
        formatter(None, **kwargs)
        return lambda prog: formatter(prog, **kwargs)
    except TypeError:
        return formatter


def fmutool():
    print(f"FMU Manipulation Toolbox version {version}")
    help_message = Help()

    parser = argparse.ArgumentParser(prog='fmutool',
                                     description="Analyse and Manipulate a FMU by modifying its 'modelDescription.xml'",
                                     formatter_class=make_wide(argparse.HelpFormatter),
                                     add_help=False,
                                     epilog="see: https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/README.md")

    def add_option(option, *args, **kwargs):
        parser.add_argument(option, *args, help=help_message.usage(option), **kwargs)

    add_option('-h', '-help', action="help")

    # I/O
    add_option('-input', action='store', dest='fmu_input', default=None, required=True, metavar='path/to/module.fmu')
    add_option('-output', action='store', dest='fmu_output', default=None, metavar='path/to/module-modified.fmu')

    # Port name manipulation
    add_option('-remove-toplevel', action='append_const', dest='operations_list', const=OperationStripTopLevel())
    add_option('-merge-toplevel', action='append_const', dest='operations_list', const=OperationMergeTopLevel())
    add_option('-trim-until', action='append', dest='operations_list', type=OperationTrimUntil, metavar='prefix')
    add_option('-remove-regexp', action='append', dest='operations_list', type=OperationRemoveRegexp,
               metavar='regular-expression')
    add_option('-keep-only-regexp', action='append', dest='operations_list', type=OperationKeepOnlyRegexp,
               metavar='regular-expression')
    add_option('-remove-all', action='append_const', dest='operations_list', const=OperationRemoveRegexp('.*'))

    # Batch Rename
    add_option('-dump-csv', action='append', dest='operations_list', type=OperationSaveNamesToCSV,
               metavar='path/to/list.csv')
    add_option('-rename-from-csv', action='append', dest='operations_list', type=OperationRenameFromCSV,
               metavar='path/to/translation.csv')

    # Remoting
    add_option('-add-remoting-win32', action='append_const', dest='operations_list', const=OperationAddRemotingWin32())
    add_option('-add-remoting-win64', action='append_const', dest='operations_list', const=OperationAddRemotingWin64())
    add_option('-add-frontend-win32', action='append_const', dest='operations_list', const=OperationAddFrontendWin32())
    add_option('-add-frontend-win64', action='append_const', dest='operations_list', const=OperationAddFrontendWin64())

    # Extraction / Removal
    add_option('-extract-descriptor', action='store', dest='extract_description',
               metavar='path/to/saved-modelDescriptor.xml')
    add_option('-remove-sources', action='append_const', dest='operations_list',
               const=OperationRemoveSources())
    # Filter
    add_option('-only-parameters', action='append_const', dest='apply_on', const='parameter')
    add_option('-only-inputs', action='append_const', dest='apply_on', const='input')
    add_option('-only-outputs', action='append_const', dest='apply_on', const='output')
    # Checker
    add_option('-summary', action='append_const', dest='operations_list', const=OperationSummary())
    add_option('-check', action='append_const', dest='operations_list', const=[checker() for checker in checker_list])

    cli_options = parser.parse_args()
    # handle the "no operation" use case
    if not cli_options.operations_list:
        cli_options.operations_list = []

    if cli_options.fmu_input == cli_options.fmu_output:
        print(f"FATAL ERROR: '-input' and '-output' should point to different files.")
        sys.exit(-3)

    print(f"READING Input='{cli_options.fmu_input}'")
    try:
        fmu = FMU(cli_options.fmu_input)
    except FMUException as reason:
        print(f"FATAL ERROR: {reason}")
        sys.exit(-4)

    if cli_options.apply_on:
        print("Applying operation for :")
        for causality in cli_options.apply_on:
            print(f"     - causality = {causality}")

    def flatten(list_of_list: list):
        return [x for xs in list_of_list for x in xs]

    for operation in flatten(cli_options.operations_list):
        print(f"     => {operation}")
        try:
            fmu.apply_operation(operation, cli_options.apply_on)
        except OperationException as reason:
            print(f"ERROR: {reason}")
            sys.exit(-6)

    if cli_options.extract_description:
        print(f"WRITING ModelDescriptor='{cli_options.extract_description}'")
        fmu.save_descriptor(cli_options.extract_description)

    if cli_options.fmu_output:
        print(f"WRITING Output='{cli_options.fmu_output}'")
        try:
            fmu.repack(cli_options.fmu_output)
        except FMUException as reason:
            print(f"FATAL ERROR: {reason}")
            sys.exit(-5)
    else:
        print(f"INFO    Modified FMU is not saved. If necessary use '-output' option.")


def fmucontainer():
    logger = setup_logger()

    logger.info(f"FMUContainer version {version}")
    parser = argparse.ArgumentParser(prog="fmucontainer", description="Generate FMU from FMU's",
                                     formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter),
                                     add_help=False,
                                     epilog="see: https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/"
                                            "container/README.md")

    parser.add_argument('-h', '-help', action="help")

    parser.add_argument("-fmu-directory", action="store", dest="fmu_directory", required=False, default=".",
                        help="Directory containing initial FMU’s and used to generate containers. "
                             "If not defined, current directory is used.")

    parser.add_argument("-container", action="append", dest="container_descriptions_list", default=[],
                        metavar="filename.{csv|json|ssp},[:step_size]", required=True,
                        help="Description of the container to create.")

    parser.add_argument("-debug", action="store_true", dest="debug",
                        help="Add lot of useful log during the process.")

    parser.add_argument("-no-auto-input", action="store_false", dest="auto_input", default=True,
                        help="Create ONLY explicit input.")

    parser.add_argument("-no-auto-output", action="store_false", dest="auto_output", default=True,
                        help="Create ONLY explicit output.")

    parser.add_argument("-auto-parameter", action="store_true", dest="auto_parameter", default=False,
                        help="Expose parameters of the embedded fmu's.")

    parser.add_argument("-auto-local", action="store_true", dest="auto_local", default=False,
                        help="Expose local variables of the embedded fmu's.")

    parser.add_argument("-no-auto-link", action="store_false", dest="auto_link", default=True,
                        help="Create ONLY explicit links.")

    parser.add_argument("-mt", action="store_true", dest="mt", default=False,
                        help="Enable Multi-Threaded mode for the generated container.")

    parser.add_argument("-profile", action="store_true", dest="profiling", default=False,
                        help="Enable Profiling mode for the generated container.")

    parser.add_argument("-dump-json",  action="store_true", dest="dump", default=False,
                        help="Dump a JSON file for each container.")

    config = parser.parse_args()

    if config.debug:
        logger.setLevel(logging.DEBUG)

    fmu_directory = Path(config.fmu_directory)
    logger.info(f"FMU directory: '{fmu_directory}'")

    for description in config.container_descriptions_list:
        try:
            tokens = description.split(":")
            filename = ":".join(tokens[:-1])
            step_size = float(tokens[-1])
        except ValueError:
            step_size = None
            filename = description

        try:
            assembly = Assembly(filename, step_size=step_size, auto_link=config.auto_link,
                                auto_input=config.auto_input, auto_output=config.auto_output,
                                auto_local=config.auto_local, mt=config.mt,
                                profiling=config.profiling, fmu_directory=fmu_directory, debug=config.debug,
                                auto_parameter=config.auto_parameter)
        except FileNotFoundError as e:
            logger.fatal(f"Cannot read file: {e}")
            continue
        except (FMUContainerError, AssemblyError) as e:
            logger.fatal(f"{filename}: {e}")
            continue

        try:
            assembly.make_fmu(dump_json=config.dump)
        except FMUContainerError as e:
            logger.fatal(f"{filename}: {e}")
            continue


if __name__ == "__main__":
    fmucontainer()



import logging
import os
import shutil
import uuid
import zipfile
from datetime import datetime
from pathlib import Path
from typing import *

from .fmu_operations import FMU, OperationAbstract
from .version import __version__ as tool_version


logger = logging.getLogger("fmu_manipulation_toolbox")


class FMUPort:
    def __init__(self, attrs: Dict[str, str]):
        self.name = attrs["name"]
        self.vr = int(attrs["valueReference"])
        self.causality = attrs.get("causality", "local")
        self.attrs = attrs.copy()
        self.attrs.pop("name")
        self.attrs.pop("valueReference")
        if "causality" in self.attrs:
            self.attrs.pop("causality")
        self.type_name = None
        self.child = None

    def set_port_type(self, type_name: str, attrs: Dict[str, str]):
        self.type_name = type_name
        self.child = attrs.copy()
        for unsupported in ("unit", "declaredType"):
            if unsupported in self.child:
                self.child.pop(unsupported)

    def xml(self, vr: int, name=None, causality=None, start=None):

        if self.child is None:
            raise FMUContainerError(f"FMUPort has no child. Bug?")

        child_str = f"<{self.type_name}"
        if self.child:
            if start is not None and 'start' in self.child:
                self.child['start'] = start
            child_str += " " + " ".join([f'{key}="{value}"' for (key, value) in self.child.items()]) + "/>"
        else:
            child_str += "/>"

        if name is None:
            name = self.name
        if causality is None:
            causality = self.causality

        variability = "continuous" if self.type_name == "Real" else "discrete"

        scalar_attrs = {
            "name": name,
            "valueReference": vr,
            "causality": causality,
            "variability": variability,
        }
        scalar_attrs.update(self.attrs)

        scalar_attrs_str = " ".join([f'{key}="{value}"' for (key, value) in scalar_attrs.items()])

        return f'<ScalarVariable {scalar_attrs_str}>{child_str}</ScalarVariable>'


class EmbeddedFMU(OperationAbstract):
    capability_list = ("needsExecutionTool",
                       "canBeInstantiatedOnlyOncePerProcess")

    def __init__(self, filename):
        self.fmu = FMU(filename)
        self.name = Path(filename).name
        self.id = Path(filename).stem

        self.fmi_version = None
        self.step_size = None
        self.start_time = None
        self.stop_time = None
        self.model_identifier = None
        self.guid = None
        self.ports: Dict[str, FMUPort] = {}

        self.capabilities: Dict[str, str] = {}
        self.current_port = None  # used during apply_operation()

        self.fmu.apply_operation(self)  # Should be the last command in constructor!
        if self.model_identifier is None:
            raise FMUContainerError(f"FMU '{self.name}' does not implement Co-Simulation mode.")

    def fmi_attrs(self, attrs):
        self.guid = attrs['guid']
        self.fmi_version = attrs['fmiVersion']

    def scalar_attrs(self, attrs) -> int:
        self.current_port = FMUPort(attrs)
        self.ports[self.current_port.name] = self.current_port

        return 0

    def cosimulation_attrs(self, attrs: Dict[str, str]):
        self.model_identifier = attrs['modelIdentifier']
        for capability in self.capability_list:
            self.capabilities[capability] = attrs.get(capability, "false")

    def experiment_attrs(self, attrs: Dict[str, str]):
        try:
            self.step_size = float(attrs['stepSize'])
        except KeyError:
            logger.warning(f"FMU '{self.name}' does not specify preferred step size")
        self.start_time = float(attrs.get("startTime", 0.0))
        self.stop_time = float(attrs.get("stopTime", self.start_time + 1.0))

    def scalar_type(self, type_name, attrs):
        if self.current_port:
            if type_name == "Enumeration":
                type_name = "Integer"
            self.current_port.set_port_type(type_name, attrs)
        self.current_port = None

    def __repr__(self):
        return f"FMU '{self.name}' ({len(self.ports)} variables)"


class FMUContainerError(Exception):
    def __init__(self, reason: str):
        self.reason = reason

    def __repr__(self):
        return f"{self.reason}"


class ContainerPort:
    def __init__(self, fmu: EmbeddedFMU, port_name: str):
        self.fmu = fmu
        try:
            self.port = fmu.ports[port_name]
        except KeyError:
            raise FMUContainerError(f"Port '{fmu.name}/{port_name}' does not exist")
        self.vr = None

    def __repr__(self):
        return f"Port {self.fmu.name}/{self.port.name}"

    def __hash__(self):
        return hash("{self.fmu.name}/{self.port.name}")

    def __eq__(self, other):
        return str(self) == str(other)


class Local:
    def __init__(self, cport_from: ContainerPort):
        self.name = cport_from.fmu.id + "." + cport_from.port.name  # strip .fmu suffix
        self.cport_from = cport_from
        self.cport_to_list: List[ContainerPort] = []
        self.vr = None

        if not cport_from.port.causality == "output":
            raise FMUContainerError(f"{cport_from} is  {cport_from.port.causality} instead of OUTPUT")

    def add_target(self, cport_to: ContainerPort):
        if not cport_to.port.causality == "input":
            raise FMUContainerError(f"{cport_to} is {cport_to.port.causality} instead of INPUT")

        if cport_to.port.type_name == self.cport_from.port.type_name:
            self.cport_to_list.append(cport_to)
        else:
            raise FMUContainerError(f"failed to connect {self.cport_from} to {cport_to} due to type.")


class ValueReferenceTable:
    def __init__(self):
        self.vr_table: Dict[str, int] = {
            "Real": 0,
            "Integer": 0,
            "Boolean": 0,
            "String": 0,
        }

    def get_vr(self, cport: ContainerPort) -> int:
        return self.add_vr(cport.port.type_name)

    def add_vr(self, type_name: str) -> int:
        vr = self.vr_table[type_name]
        self.vr_table[type_name] += 1
        return vr


class FMUContainer:
    def __init__(self, identifier: str, fmu_directory: Union[str, Path], description_pathname=None):
        self.fmu_directory = Path(fmu_directory)
        self.identifier = identifier
        if not self.fmu_directory.is_dir():
            raise FMUContainerError(f"{self.fmu_directory} is not a valid directory")
        self.involved_fmu: Dict[str, EmbeddedFMU] = {}
        self.execution_order: List[EmbeddedFMU] = []

        self.description_pathname = description_pathname

        self.start_time = None
        self.stop_time = None

        # Rules
        self.inputs: Dict[str, ContainerPort] = {}
        self.outputs: Dict[str, ContainerPort] = {}
        self.locals: Dict[ContainerPort, Local] = {}

        self.rules: Dict[ContainerPort, str] = {}
        self.start_values: Dict[ContainerPort, str] = {}

    def get_fmu(self, fmu_filename: str) -> EmbeddedFMU:
        if fmu_filename in self.involved_fmu:
            return self.involved_fmu[fmu_filename]

        try:
            fmu = EmbeddedFMU(self.fmu_directory / fmu_filename)
            self.involved_fmu[fmu_filename] = fmu
            self.execution_order.append(fmu)
            if not fmu.fmi_version == "2.0":
                raise FMUContainerError("Only FMI-2.0 is supported by FMUContainer")
            logger.debug(f"Adding FMU #{len(self.execution_order)}: {fmu}")
        except Exception as e:
            raise FMUContainerError(f"Cannot load '{fmu_filename}': {e}")

        return fmu

    def mark_ruled(self, cport: ContainerPort, rule: str):
        if cport in self.rules:
            previous_rule = self.rules[cport]
            if rule not in ("OUTPUT", "LINK") and previous_rule not in ("OUTPUT", "LINK"):
                raise FMUContainerError(f"try to {rule} port {cport} which is already {previous_rule}")

        self.rules[cport] = rule

    def add_input(self, container_port_name: str, to_fmu_filename: str, to_port_name: str):
        if not container_port_name:
            container_port_name = to_port_name
        cport_to = ContainerPort(self.get_fmu(to_fmu_filename), to_port_name)
        if not cport_to.port.causality == "input":  # check causality
            raise FMUContainerError(f"Tried to use '{cport_to}' as INPUT of the container but FMU causality is "
                                    f"'{cport_to.port.causality}'.")

        logger.debug(f"INPUT: {to_fmu_filename}:{to_port_name}")
        self.mark_ruled(cport_to, 'INPUT')
        self.inputs[container_port_name] = cport_to

    def add_output(self, from_fmu_filename: str, from_port_name: str, container_port_name: str):
        if not container_port_name:  # empty is allowed
            container_port_name = from_port_name

        cport_from = ContainerPort(self.get_fmu(from_fmu_filename), from_port_name)
        if cport_from.port.causality not in ("output", "local"):  # check causality
            raise FMUContainerError(f"Tried to use '{cport_from}' as INPUT of the container but FMU causality is "
                                    f"'{cport_from.port.causality}'.")

        logger.debug(f"OUTPUT: {from_fmu_filename}:{from_port_name}")
        self.mark_ruled(cport_from, 'OUTPUT')
        self.outputs[container_port_name] = cport_from

    def drop_port(self, from_fmu_filename: str, from_port_name: str):
        cport_from = ContainerPort(self.get_fmu(from_fmu_filename), from_port_name)
        if not cport_from.port.causality == "output":  # check causality
            raise FMUContainerError(f"{cport_from}: trying to DROP {cport_from.port.causality}")

        logger.debug(f"DROP: {from_fmu_filename}:{from_port_name}")
        self.mark_ruled(cport_from, 'DROP')

    def add_link(self, from_fmu_filename: str, from_port_name: str, to_fmu_filename: str, to_port_name: str):
        cport_from = ContainerPort(self.get_fmu(from_fmu_filename), from_port_name)
        try:
            local = self.locals[cport_from]
        except KeyError:
            local = Local(cport_from)

        cport_to = ContainerPort(self.get_fmu(to_fmu_filename), to_port_name)
        local.add_target(cport_to)  # Causality is check in the add() function

        self.mark_ruled(cport_from, 'LINK')
        self.mark_ruled(cport_to, 'LINK')
        self.locals[cport_from] = local

    def add_start_value(self, fmu_filename: str, port_name: str, value: str):
        cport = ContainerPort(self.get_fmu(fmu_filename), port_name)

        try:
            if cport.port.type_name == 'Real':
                value = float(value)
            elif cport.port.type_name == 'Integer':
                value = int(value)
            elif cport.port.type_name == 'Boolean':
                value = int(bool(value))
            else:
                value = value
        except ValueError:
            raise FMUContainerError(f"Start value is not conforming to '{cport.port.type_name}' format.")

        self.start_values[cport] = value

    def find_inputs(self, port_to_connect: FMUPort) -> List[ContainerPort]:
        candidates = []
        for fmu in self.execution_order:
            for port in fmu.ports.values():
                if (port.causality == 'input' and port.name == port_to_connect.name
                        and port.type_name == port_to_connect.type_name):
                    candidates.append(ContainerPort(fmu, port.name))
        return candidates

    def add_implicit_rule(self, auto_input=True, auto_output=True, auto_link=True, auto_parameter=False,
                          auto_local=False):
        # Auto Link outputs
        for fmu in self.execution_order:
            for port_name in fmu.ports:
                cport = ContainerPort(fmu, port_name)
                if cport not in self.rules:
                    if cport.port.causality == 'parameter' and auto_parameter:
                        parameter_name = cport.fmu.id + "." + cport.port.name
                        logger.info(f"AUTO PARAMETER: {cport} as {parameter_name}")
                        self.inputs[parameter_name] = cport
                        self.mark_ruled(cport, 'PARAMETER')
                    elif cport.port.causality == 'output':
                        candidates_cport_list = self.find_inputs(cport.port)
                        if auto_link and candidates_cport_list:
                            self.locals[cport] = Local(cport)
                            self.mark_ruled(cport, 'LINK')
                            for candidate_cport in candidates_cport_list:
                                self.locals[cport].add_target(candidate_cport)
                                self.mark_ruled(candidate_cport, 'LINK')
                                logger.info(f"AUTO LINK: {cport} -> {candidate_cport}")
                        elif auto_output:
                            self.mark_ruled(cport, 'OUTPUT')
                            self.outputs[port_name] = cport
                            logger.info(f"AUTO OUTPUT: Expose {cport}")
                    elif cport.port.causality == 'local':
                        local_portname = None
                        if port_name.startswith("container."):
                            local_portname = "container." + cport.fmu.id + "." + port_name[10:]
                            logger.info(f"PROFILING: Expose {cport}")
                        elif auto_local:
                            local_portname = cport.fmu.id + "." + port_name
                            logger.info(f"AUTO LOCAL: Expose {cport}")
                        if local_portname:
                            self.mark_ruled(cport, 'OUTPUT')
                            self.outputs[local_portname] = cport

        if auto_input:
            # Auto link inputs
            for fmu in self.execution_order:
                for port_name in fmu.ports:
                    cport = ContainerPort(fmu, port_name)
                    if cport not in self.rules:
                        if cport.port.causality == 'input':
                            self.mark_ruled(cport, 'INPUT')
                            self.inputs[port_name] = cport
                            logger.info(f"AUTO INPUT: Expose {cport}")

    def minimum_step_size(self) -> float:
        step_size = None
        for fmu in self.execution_order:
            if step_size:
                if fmu.step_size and fmu.step_size < step_size:
                    step_size = fmu.step_size
            else:
                step_size = fmu.step_size

        if not step_size:
            step_size = 0.1
            logger.warning(f"Defaulting to step_size={step_size}")

        return step_size

    def sanity_check(self, step_size: Optional[float]):
        for fmu in self.execution_order:
            if not fmu.step_size:
                continue
            ts_ratio = step_size / fmu.step_size
            if ts_ratio < 1.0:
                logger.warning(f"Container step_size={step_size}s is lower than FMU '{fmu.name}' "
                               f"step_size={fmu.step_size}s.")
            if ts_ratio != int(ts_ratio):
                logger.warning(f"Container step_size={step_size}s should divisible by FMU '{fmu.name}' "
                               f"step_size={fmu.step_size}s.")
            for port_name in fmu.ports:
                cport = ContainerPort(fmu, port_name)
                if cport not in self.rules:
                    if cport.port.causality == 'input':
                        logger.error(f"{cport} is not connected")
                    if cport.port.causality == 'output':
                        logger.warning(f"{cport} is not connected")

    def make_fmu(self, fmu_filename: Union[str, Path], step_size: Optional[float] = None, debug=False, mt=False,
                 profiling=False):
        if isinstance(fmu_filename, str):
            fmu_filename = Path(fmu_filename)

        if step_size is None:
            logger.info(f"step_size  will be deduced from the embedded FMU's")
            step_size = self.minimum_step_size()
        self.sanity_check(step_size)

        logger.info(f"Building FMU '{fmu_filename}', step_size={step_size}")

        base_directory = self.fmu_directory / fmu_filename.with_suffix('')
        resources_directory = self.make_fmu_skeleton(base_directory)
        with open(base_directory / "modelDescription.xml", "wt") as xml_file:
            self.make_fmu_xml(xml_file, step_size, profiling)
        with open(resources_directory / "container.txt", "wt") as txt_file:
            self.make_fmu_txt(txt_file, step_size, mt, profiling)

        self.make_fmu_package(base_directory, fmu_filename)
        if not debug:
            self.make_fmu_cleanup(base_directory)

    def make_fmu_xml(self, xml_file, step_size: float, profiling: bool):
        vr_table = ValueReferenceTable()

        timestamp = datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        guid = str(uuid.uuid4())
        embedded_fmu = ", ".join([fmu_name for fmu_name in self.involved_fmu])
        try:
            author = os.getlogin()
        except OSError:
            author = "Unspecified"

        capabilities = {}
        for capability in EmbeddedFMU.capability_list:
            capabilities[capability] = "false"
            for fmu in self.involved_fmu.values():
                if fmu.capabilities[capability] == "true":
                    capabilities[capability] = "true"

        if self.start_time is None:
            self.start_time = self.execution_order[0].start_time
            logger.info(f"start_time={self.start_time} (deduced from '{self.execution_order[0].name}')")
        else:
            logger.info(f"start_time={self.start_time}")

        if self.stop_time is None:
            self.stop_time = self.execution_order[0].stop_time
            logger.info(f"stop_time={self.stop_time} (deduced from '{self.execution_order[0].name}')")
        else:
            logger.info(f"stop_time={self.stop_time}")

        xml_file.write(f"""<?xml version="1.0" encoding="ISO-8859-1"?>
<fmiModelDescription
  fmiVersion="2.0"
  modelName="{self.identifier}"
  generationTool="FMUContainer-{tool_version}"
  generationDateAndTime="{timestamp}"
  guid="{guid}"
  description="FMUContainer with {embedded_fmu}"
  author="{author}"
  license="Proprietary"
  copyright="See Embedded FMU's copyrights."
  variableNamingConvention="structured">

  <CoSimulation
    modelIdentifier="{self.identifier}"
    canHandleVariableCommunicationStepSize="true"
    canBeInstantiatedOnlyOncePerProcess="{capabilities['canBeInstantiatedOnlyOncePerProcess']}"
    canNotUseMemoryManagementFunctions="true"
    canGetAndSetFMUstate="false"
    canSerializeFMUstate="false"
    providesDirectionalDerivative="false"
    needsExecutionTool="{capabilities['needsExecutionTool']}">
  </CoSimulation>

  <LogCategories>
    <Category name="fmucontainer"/>
  </LogCategories>

  <DefaultExperiment stepSize="{step_size}" startTime="{self.start_time}" stopTime="{self.stop_time}"/>

  <ModelVariables>
""")
        if profiling:
            for fmu in self.execution_order:
                vr = vr_table.add_vr("Real")
                name = f"container.{fmu.id}.rt_ratio"
                print(f'<ScalarVariable valueReference="{vr}" name="{name}" causality="local">'
                      f'<Real /></ScalarVariable>', file=xml_file)

        # Local variable should be first to ensure to attribute them the lowest VR.
        for local in self.locals.values():
            vr = vr_table.get_vr(local.cport_from)
            print(f'    {local.cport_from.port.xml(vr, name=local.name, causality="local")}', file=xml_file)
            local.vr = vr

        for input_port_name, cport in self.inputs.items():
            vr = vr_table.get_vr(cport)
            start = self.start_values.get(cport, None)
            print(f"    {cport.port.xml(vr, name=input_port_name, start=start)}", file=xml_file)
            cport.vr = vr

        for output_port_name, cport in self.outputs.items():
            vr = vr_table.get_vr(cport)
            print(f"    {cport.port.xml(vr, name=output_port_name)}", file=xml_file)
            cport.vr = vr

        xml_file.write("""  </ModelVariables>

  <ModelStructure>
    <Outputs>
""")

        index_offset = len(self.locals) + len(self.inputs) + 1
        for i, _ in enumerate(self.outputs.keys()):
            print(f'      <Unknown index="{index_offset+i}"/>', file=xml_file)
        xml_file.write("""    </Outputs>
    <InitialUnknowns>
""")
        for i, _ in enumerate(self.outputs.keys()):
            print(f'      <Unknown index="{index_offset+i}"/>', file=xml_file)
        xml_file.write("""    </InitialUnknowns>
  </ModelStructure>

</fmiModelDescription>
""")

    def make_fmu_txt(self, txt_file, step_size: float, mt: bool, profiling: bool):
        if mt:
            print("# Use MT\n1", file=txt_file)
        else:
            print("# Don't use MT\n0", file=txt_file)

        if profiling:
            print("# Profiling ENABLED\n1", file=txt_file)
        else:
            print("# Profiling DISABLED\n0", file=txt_file)

        print(f"# Internal time step in seconds", file=txt_file)
        print(f"{step_size}", file=txt_file)
        print(f"# NB of embedded FMU's", file=txt_file)
        print(f"{len(self.involved_fmu)}", file=txt_file)
        fmu_rank: Dict[str, int] = {}
        for i, fmu in enumerate(self.execution_order):
            print(f"{fmu.name}", file=txt_file)
            print(f"{fmu.model_identifier}", file=txt_file)
            print(f"{fmu.guid}", file=txt_file)
            fmu_rank[fmu.name] = i

        # Prepare data structure
        type_names_list = ("Real", "Integer", "Boolean", "String")  # Ordered list
        inputs_per_type: Dict[str, List[ContainerPort]] = {}        # Container's INPUT
        outputs_per_type: Dict[str, List[ContainerPort]] = {}       # Container's OUTPUT

        inputs_fmu_per_type: Dict[str, Dict[str, Dict[ContainerPort, int]]] = {}      # [type][fmu]
        start_values_fmu_per_type = {}
        outputs_fmu_per_type = {}
        locals_per_type: Dict[str, List[Local]] = {}

        for type_name in type_names_list:
            inputs_per_type[type_name] = []
            outputs_per_type[type_name] = []
            locals_per_type[type_name] = []

            inputs_fmu_per_type[type_name] = {}
            start_values_fmu_per_type[type_name] = {}
            outputs_fmu_per_type[type_name] = {}

            for fmu in self.execution_order:
                inputs_fmu_per_type[type_name][fmu.name] = {}
                start_values_fmu_per_type[type_name][fmu.name] = {}
                outputs_fmu_per_type[type_name][fmu.name] = {}

        # Fill data structure
        # Inputs
        for input_port_name, cport in self.inputs.items():
            inputs_per_type[cport.port.type_name].append(cport)
        for cport, value in self.start_values.items():
            start_values_fmu_per_type[cport.port.type_name][cport.fmu.name][cport] = value
        # Outputs
        for output_port_name, cport in self.outputs.items():
            outputs_per_type[cport.port.type_name].append(cport)
        # Locals
        for local in self.locals.values():
            vr = local.vr
            locals_per_type[local.cport_from.port.type_name].append(local)
            outputs_fmu_per_type[local.cport_from.port.type_name][local.cport_from.fmu.name][local.cport_from] = vr
            for cport_to in local.cport_to_list:
                inputs_fmu_per_type[cport_to.port.type_name][cport_to.fmu.name][cport_to] = vr

        print(f"# NB local variables Real, Integer, Boolean, String", file=txt_file)
        for type_name in type_names_list:
            nb = len(locals_per_type[type_name])
            if profiling and type_name == "Real":
                nb += len(self.execution_order)
            print(f"{nb} ", file=txt_file, end='')
        print("", file=txt_file)

        print("# CONTAINER I/O: <VR> <FMU_INDEX> <FMU_VR>", file=txt_file)
        for type_name in type_names_list:
            print(f"# {type_name}", file=txt_file)
            nb = len(inputs_per_type[type_name])+len(outputs_per_type[type_name])+len(locals_per_type[type_name])
            if profiling and type_name == "Real":
                nb += len(self.execution_order)
                print(nb, file=txt_file)
                for profiling_port, _ in enumerate(self.execution_order):
                    print(f"{profiling_port} -2 {profiling_port}", file=txt_file)
            else:
                print(nb, file=txt_file)
            for cport in inputs_per_type[type_name]:
                print(f"{cport.vr} {fmu_rank[cport.fmu.name]} {cport.port.vr}", file=txt_file)
            for cport in outputs_per_type[type_name]:
                print(f"{cport.vr} {fmu_rank[cport.fmu.name]} {cport.port.vr}", file=txt_file)
            for local in locals_per_type[type_name]:
                print(f"{local.vr} -1 {local.vr}", file=txt_file)

        # LINKS
        for fmu in self.execution_order:
            for type_name in type_names_list:
                print(f"# Inputs of {fmu.name} - {type_name}: <VR> <FMU_VR>", file=txt_file)
                print(len(inputs_fmu_per_type[type_name][fmu.name]), file=txt_file)
                for cport, vr in inputs_fmu_per_type[type_name][fmu.name].items():
                    print(f"{vr} {cport.port.vr}", file=txt_file)

            for type_name in type_names_list:
                print(f"# Start values of {fmu.name} - {type_name}: <FMU_VR> <VALUE>", file=txt_file)
                print(len(start_values_fmu_per_type[type_name][fmu.name]), file=txt_file)
                for cport, value in start_values_fmu_per_type[type_name][fmu.name].items():
                    print(f"{cport.port.vr} {value}", file=txt_file)

            for type_name in type_names_list:
                print(f"# Outputs of {fmu.name} - {type_name}: <VR> <FMU_VR>", file=txt_file)
                print(len(outputs_fmu_per_type[type_name][fmu.name]), file=txt_file)
                for cport, vr in outputs_fmu_per_type[type_name][fmu.name].items():
                    print(f"{vr} {cport.port.vr}", file=txt_file)

    @staticmethod
    def long_path(path: Union[str, Path]) -> str:
        # https://stackoverflow.com/questions/14075465/copy-a-file-with-a-too-long-path-to-another-directory-in-python
        if os.name == 'nt':
            return "\\\\?\\" + os.path.abspath(str(path))
        else:
            return path

    def make_fmu_skeleton(self, base_directory: Path) -> Path:
        logger.debug(f"Initialize directory '{base_directory}'")

        origin = Path(__file__).parent / "resources"
        resources_directory = base_directory / "resources"
        documentation_directory = base_directory / "documentation"
        binaries_directory = base_directory / "binaries"

        base_directory.mkdir(exist_ok=True)
        resources_directory.mkdir(exist_ok=True)
        binaries_directory.mkdir(exist_ok=True)
        documentation_directory.mkdir(exist_ok=True)

        if self.description_pathname:
            logger.debug(f"Copying {self.description_pathname}")
            shutil.copy(self.description_pathname, documentation_directory)

        shutil.copy(origin / "model.png", base_directory)
        for bitness in ('win32', 'win64'):
            library_filename = origin / bitness / "container.dll"
            if library_filename.is_file():
                binary_directory = binaries_directory / bitness
                binary_directory.mkdir(exist_ok=True)
                shutil.copy(library_filename, binary_directory / f"{self.identifier}.dll")

        for fmu in self.involved_fmu.values():
            shutil.copytree(self.long_path(fmu.fmu.tmp_directory),
                            self.long_path(resources_directory / fmu.name), dirs_exist_ok=True)
        return resources_directory

    def make_fmu_package(self, base_directory: Path, fmu_filename: Path):
        logger.debug(f"Zipping directory '{base_directory}' => '{fmu_filename}'")
        zip_directory = self.long_path(str(base_directory.absolute()))
        offset = len(zip_directory) + 1
        with zipfile.ZipFile(self.fmu_directory / fmu_filename, "w", zipfile.ZIP_DEFLATED) as zip_file:
            def add_file(directory: Path):
                for entry in directory.iterdir():
                    if entry.is_dir():
                        add_file(directory / entry)
                    elif entry.is_file:
                        zip_file.write(str(entry), str(entry)[offset:])

            add_file(Path(zip_directory))
        logger.info(f"'{fmu_filename}' is available.")

    def make_fmu_cleanup(self, base_directory: Path):
        logger.debug(f"Delete directory '{base_directory}'")
        shutil.rmtree(self.long_path(base_directory))



import csv
import json
import logging
from typing import *
from pathlib import Path
import uuid
import xml.parsers.expat
import zipfile

from .fmu_container import FMUContainer

logger = logging.getLogger("fmu_manipulation_toolbox")


class Port:
    def __init__(self, fmu_name: str, port_name: str):
        self.fmu_name = fmu_name
        self.port_name = port_name

    def __hash__(self):
        return hash(f"{self.fmu_name}/{self.port_name}")

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return f"{self.fmu_name}/{self.port_name}"


class Connection:
    def __init__(self, from_port: Port, to_port: Port):
        self.from_port = from_port
        self.to_port = to_port

    def __str__(self):
        return f"{self.from_port} -> {self.to_port}"


class AssemblyNode:
    def __init__(self, name: Optional[str], step_size: float = None, mt=False, profiling=False,
                 auto_link=True, auto_input=True, auto_output=True, auto_parameter=False, auto_local=False):
        self.name = name
        if step_size:
            try:
                self.step_size = float(step_size)
            except ValueError:
                logger.warning(f"Step size '{step_size}' is incorrect format.")
                self.step_size = None
        else:
            self.step_size = None
        self.mt = mt
        self.profiling = profiling
        self.auto_link = auto_link
        self.auto_input = auto_input
        self.auto_output = auto_output
        self.auto_parameter = auto_parameter
        self.auto_local = auto_local

        self.parent: Optional[AssemblyNode] = None
        self.children: Dict[str, AssemblyNode] = {}     # sub-containers
        self.fmu_names_list: List[str] = []             # FMUs contained at this level (ordered list)
        self.input_ports: Dict[Port, str] = {}          # value is input port name, key is the source
        self.output_ports: Dict[Port, str] = {}         # value is output port name, key is the origin
        self.start_values: Dict[Port, str] = {}
        self.drop_ports: List[Port] = []
        self.links: List[Connection] = []

    def add_sub_node(self, sub_node):
        if sub_node.name is None:
            sub_node.name = str(uuid.uuid4())+".fmu"

        if sub_node.parent is not None:
            raise AssemblyError(f"Internal Error: AssemblyNode {sub_node.name} is already parented.")

        if sub_node.name in self.children:
            raise AssemblyError(f"Internal Error: AssemblyNode {sub_node.name} is already child of {self.name}")

        sub_node.parent = self
        if sub_node.name not in self.fmu_names_list:
            self.fmu_names_list.append(sub_node.name)
        self.children[sub_node.name] = sub_node

    def add_fmu(self, fmu_name: str):
        if fmu_name not in self.fmu_names_list:
            self.fmu_names_list.append(fmu_name)

    def add_input(self, from_port_name: str, to_fmu_filename: str, to_port_name: str):
        self.input_ports[Port(to_fmu_filename, to_port_name)] = from_port_name

    def add_output(self, from_fmu_filename: str, from_port_name: str, to_port_name: str):
        self.output_ports[Port(from_fmu_filename, from_port_name)] = to_port_name

    def add_drop_port(self, fmu_filename: str, port_name: str):
        self.drop_ports.append(Port(fmu_filename, port_name))

    def add_link(self, from_fmu_filename: str, from_port_name: str, to_fmu_filename: str, to_port_name: str):
        self.links.append(Connection(Port(from_fmu_filename, from_port_name),
                          Port(to_fmu_filename, to_port_name)))

    def add_start_value(self, fmu_filename: str, port_name: str, value: str):
        self.start_values[Port(fmu_filename, port_name)] = value

    def make_fmu(self, fmu_directory: Path, debug=False, description_pathname=None):
        for node in self.children.values():
            node.make_fmu(fmu_directory, debug=debug)

        identifier = str(Path(self.name).stem)
        container = FMUContainer(identifier, fmu_directory, description_pathname=description_pathname)

        for fmu_name in self.fmu_names_list:
            container.get_fmu(fmu_name)

        for port, source in self.input_ports.items():
            container.add_input(source, port.fmu_name, port.port_name)

        for port, target in self.output_ports.items():
            container.add_output(port.fmu_name, port.port_name, target)

        for link in self.links:
            container.add_link(link.from_port.fmu_name, link.from_port.port_name,
                               link.to_port.fmu_name, link.to_port.port_name)

        for drop in self.drop_ports:
            container.drop_port(drop.fmu_name, drop.port_name)

        for port, value in self.start_values.items():
            container.add_start_value(port.fmu_name, port.port_name, value)

        container.add_implicit_rule(auto_input=self.auto_input,
                                    auto_output=self.auto_output,
                                    auto_link=self.auto_link,
                                    auto_parameter=self.auto_parameter,
                                    auto_local=self.auto_local)

        container.make_fmu(self.name, self.step_size, mt=self.mt, profiling=self.profiling, debug=debug)

        for node in self.children.values():
            logger.info(f"Deleting transient FMU Container '{node.name}'")
            (fmu_directory / node.name).unlink()

    def get_final_from(self, port: Port) -> Port:
        if port in self.input_ports:
            ancestor = Port(self.name, self.input_ports[port])
            if self.parent:
                return self.parent.get_final_from(ancestor)  # input port
            else:
                return ancestor  # TOPLEVEL input port
        elif port.fmu_name in self.fmu_names_list:
            if port.fmu_name in self.children:
                child = self.children[port.fmu_name]
                ancestors = [key for key, val in child.output_ports.items() if val == port.port_name]
                if len(ancestors) == 1:
                    return child.get_final_from(ancestors[0])  # child output port
            else:
                return port  # embedded FMU

        raise AssemblyError(f"{self.name}: Port {port} is not connected upstream.")

    def get_final_to(self, port: Port) -> Port:
        if port in self.output_ports:
            successor = Port(self.name, self.output_ports[port])
            if self.parent:
                return self.parent.get_final_to(successor)  # Output port
            else:
                return successor  # TOLEVEL output port
        elif port.fmu_name in self.fmu_names_list:
            if port.fmu_name in self.children:
                child = self.children[port.fmu_name]
                successors = [key for key, val in child.input_ports.items() if val == port.port_name]
                if len(successors) == 1:
                    return child.get_final_to(successors[0])  # Child input port
            else:
                return port  # embedded FMU

        raise AssemblyError(f"Node {self.name}: Port {port} is not connected downstream.")

    def get_fmu_connections(self, fmu_name: str) -> List[Connection]:
        connections = []
        if fmu_name not in self.fmu_names_list:
            raise AssemblyError(f"Internal Error: FMU {fmu_name} is not embedded by {self.name}.")
        for link in self.links:
            if link.from_port.fmu_name == fmu_name:
                connections.append(Connection(link.from_port, self.get_final_to(link.to_port)))
            elif link.to_port.fmu_name == fmu_name:
                connections.append(Connection(self.get_final_from(link.from_port), link.to_port))

        for to_port, input_port_name in self.input_ports.items():
            if to_port.fmu_name == fmu_name:
                if self.parent:
                    connections.append(Connection(self.parent.get_final_from(Port(self.name, input_port_name)), to_port))
                else:
                    connections.append(Connection(Port(self.name, input_port_name), to_port))

        for from_port, output_port_name in self.output_ports.items():
            if from_port.fmu_name == fmu_name:
                if self.parent:
                    connections.append(Connection(from_port, self.parent.get_final_to(Port(self.name, output_port_name))))
                else:
                    connections.append(Connection(from_port, Port(self.name, output_port_name))) ###HERE

        return connections


class AssemblyError(Exception):
    def __init__(self, reason: str):
        self.reason = reason

    def __repr__(self):
        return f"{self.reason}"


class Assembly:
    def __init__(self, filename: str, step_size=None, auto_link=True,  auto_input=True, debug=False,
                 auto_output=True, mt=False, profiling=False, fmu_directory: Path = Path("."), auto_parameter=False,
                 auto_local=False):
        self.filename = Path(filename)
        self.default_auto_input = auto_input
        self.debug = debug
        self.default_auto_output = auto_output
        self.default_step_size = step_size
        self.default_auto_link = auto_link
        self.default_auto_parameter = auto_parameter
        self.default_auto_local = auto_local
        self.default_mt = mt
        self.default_profiling = profiling
        self.fmu_directory = fmu_directory
        self.transient_filenames: List[Path] = []
        self.transient_dirnames: Set[Path] = set()

        if not fmu_directory.is_dir():
            raise AssemblyError(f"FMU directory is not valid: '{fmu_directory}'")

        self.input_pathname = fmu_directory / self.filename
        self.description_pathname = self.input_pathname   # For inclusion in FMU
        self.root = None
        self.read()

    def add_transient_file(self, filename: str):
        self.transient_filenames.append(self.fmu_directory / filename)
        self.transient_dirnames.add(Path(filename).parent)

    def __del__(self):
        if not self.debug:
            for filename in self.transient_filenames:
                try:
                    filename.unlink()
                except FileNotFoundError:
                    pass
            for dirname in self.transient_dirnames:
                while not str(dirname) == ".":
                    try:
                        (self.fmu_directory / dirname).rmdir()
                    except FileNotFoundError:
                        pass
                    dirname = dirname.parent

    def read(self):
        logger.info(f"Reading '{self.filename}'")
        if self.filename.suffix == ".json":
            self.read_json()
        elif self.filename.suffix == ".ssp":
            self.read_ssp()
        elif self.filename.suffix == ".csv":
            self.read_csv()
        else:
            raise AssemblyError(f"Not supported file format '{self.filename}")

    def write(self, filename: str):
        if filename.endswith(".csv"):
            return self.write_csv(filename)
        elif filename.endswith(".json"):
            return self.write_json(filename)
        else:
            logger.critical(f"Unable to write to '{filename}': format unsupported.")

    def read_csv(self):
        name = str(self.filename.with_suffix(".fmu"))
        self.root = AssemblyNode(name, step_size=self.default_step_size, auto_link=self.default_auto_link,
                                 mt=self.default_mt, profiling=self.default_profiling,
                                 auto_input=self.default_auto_input, auto_output=self.default_auto_output,
                                 auto_parameter=self.default_auto_parameter, auto_local=self.default_auto_local)

        with open(self.input_pathname) as file:
            reader = csv.reader(file, delimiter=';')
            self._check_csv_headers(reader)
            for i, row in enumerate(reader):
                if not row or row[0][0] == '#':  # skip blank line of comment
                    continue

                try:
                    rule, from_fmu_filename, from_port_name, to_fmu_filename, to_port_name = row
                except ValueError:
                    logger.error(f"Line #{i+2}: expecting 5 columns. Line skipped.")
                    continue

                try:
                    self._read_csv_rule(self.root, rule.upper(),
                                        from_fmu_filename, from_port_name, to_fmu_filename, to_port_name)
                except AssemblyError as e:
                    logger.error(f"Line #{i+2}: {e}. Line skipped.")
                    continue

    @staticmethod
    def _check_csv_headers(reader):
        headers = next(reader)
        headers_lowered = [h.lower() for h in headers]
        if not headers_lowered == ["rule", "from_fmu", "from_port", "to_fmu", "to_port"]:
            raise AssemblyError("Header (1st line of the file) is not well formatted.")

    @staticmethod
    def _read_csv_rule(node: AssemblyNode, rule: str, from_fmu_filename: str, from_port_name: str,
                       to_fmu_filename: str, to_port_name: str):
        if rule == "FMU":
            if not from_fmu_filename:
                raise AssemblyError("Missing FMU information.")
            node.add_fmu(from_fmu_filename)

        elif rule == "INPUT":
            if not to_fmu_filename or not to_port_name:
                raise AssemblyError("Missing INPUT ports information.")
            if not from_port_name:
                from_port_name = to_port_name
            node.add_input(from_port_name, to_fmu_filename, to_port_name)

        elif rule == "OUTPUT":
            if not from_fmu_filename or not from_port_name:
                raise AssemblyError("Missing OUTPUT ports information.")
            if not to_port_name:
                to_port_name = from_port_name
            node.add_output(from_fmu_filename, from_port_name, to_port_name)

        elif rule == "DROP":
            if not from_fmu_filename or not from_port_name:
                raise AssemblyError("Missing DROP ports information.")
            node.add_drop_port(from_fmu_filename, from_port_name)

        elif rule == "LINK":
            node.add_link(from_fmu_filename, from_port_name, to_fmu_filename, to_port_name)

        elif rule == "START":
            if not from_fmu_filename or not from_port_name or not to_fmu_filename:
                raise AssemblyError("Missing START ports information.")

            node.add_start_value(from_fmu_filename, from_port_name, to_fmu_filename)
        else:
            raise AssemblyError(f"unexpected rule '{rule}'. Line skipped.")

    def write_csv(self, filename: Union[str, Path]):
        if self.root.children:
            raise AssemblyError("This assembly is not flat. Cannot export to CSV file.")

        with open(self.fmu_directory / filename, "wt") as outfile:
            outfile.write("rule;from_fmu;from_port;to_fmu;to_port\n")
            for fmu in self.root.fmu_names_list:
                outfile.write(f"FMU;{fmu};;;\n")
            for port, source in self.root.input_ports.items():
                outfile.write(f"INPUT;;{source};{port.fmu_name};{port.port_name}\n")
            for port, target in self.root.output_ports.items():
                outfile.write(f"OUTPUT;{port.fmu_name};{port.port_name};;{target}\n")
            for link in self.root.links:
                outfile.write(f"LINK;{link.from_port.fmu_name};{link.from_port.port_name};"
                              f"{link.to_port.fmu_name};{link.to_port.port_name}\n")
            for port, value in self.root.start_values.items():
                outfile.write(f"START;{port.fmu_name};{port.port_name};{value};\n")
            for port in self.root.drop_ports:
                outfile.write(f"DROP;{port.fmu_name};{port.port_name};;\n")

    def read_json(self):
        with open(self.input_pathname) as file:
            try:
                data = json.load(file)
            except json.decoder.JSONDecodeError as e:
                raise AssemblyError(f"Cannot read json: {e}")
        self.root = self._json_decode_node(data)
        if not self.root.name:
            self.root.name = str(self.filename.with_suffix(".fmu").name)

    def _json_decode_node(self, data: Dict) -> AssemblyNode:
        name = data.get("name", None)                                                       # 1
        mt = data.get("mt", self.default_mt)                                                # 2
        profiling = data.get("profiling", self.default_profiling)                           # 3
        auto_link = data.get("auto_link", self.default_auto_link)                           # 4
        auto_input = data.get("auto_input", self.default_auto_input)                        # 5
        auto_output = data.get("auto_output", self.default_auto_output)                     # 6
        auto_parameter = data.get("auto_parameter", self.default_auto_parameter)            # 6b
        auto_local = data.get("auto_local", self.default_auto_local)                        # 6c
        step_size = data.get("step_size", self.default_step_size)                           # 7

        node = AssemblyNode(name, step_size=step_size, auto_link=auto_link, mt=mt, profiling=profiling,
                            auto_input=auto_input, auto_output=auto_output, auto_parameter=auto_parameter,
                            auto_local=auto_local)

        for key, value in data.items():
            if key in ('name', 'step_size', 'auto_link', 'auto_input', 'auto_output', 'mt', 'profiling',
                       'auto_parameter', 'auto_local'):
                continue  # Already read

            elif key == "container":  # 8
                if not isinstance(value, list):
                    raise AssemblyError("JSON: 'container' keyword should define a list.")
                for sub_data in value:
                    node.add_sub_node(self._json_decode_node(sub_data))

            elif key == "fmu":  # 9
                if not isinstance(value, list):
                    raise AssemblyError("JSON: 'fmu' keyword should define a list.")
                for fmu in value:
                    node.add_fmu(fmu)

            elif key == "input":  # 10
                self._json_decode_keyword('input', value, node.add_input)

            elif key == "output":  # 11
                self._json_decode_keyword('output', value, node.add_output)

            elif key == "link":  # 12
                self._json_decode_keyword('link', value, node.add_link)

            elif key == "start":  # 13
                self._json_decode_keyword('start', value, node.add_start_value)

            elif key == "drop":  # 14
                self._json_decode_keyword('drop', value, node.add_drop_port)

            else:
                logger.error(f"JSON: unexpected keyword {key}. Skipped.")

        return node

    @staticmethod
    def _json_decode_keyword(keyword: str, value, function):
        if not isinstance(value, list):
            raise AssemblyError(f"JSON: '{keyword}' keyword should define a list.")
        for line in value:
            if not isinstance(line, list):
                raise AssemblyError(f"JSON: unexpected '{keyword}' value: {line}.")
            try:
                function(*line)
            except TypeError:
                raise AssemblyError(f"JSON: '{keyword}' value does not contain right number of fields: {line}.")

    def write_json(self, filename: Union[str, Path]):
        with open(self.fmu_directory / filename, "wt") as file:
            data = self._json_encode_node(self.root)
            json.dump(data, file, indent=2)

    def _json_encode_node(self, node: AssemblyNode) -> Dict[str, Any]:
        json_node = dict()
        json_node["name"] = node.name                      # 1
        json_node["mt"] = node.mt                          # 2
        json_node["profiling"] = node.profiling            # 3
        json_node["auto_link"] = node.auto_link            # 4
        json_node["auto_input"] = node.auto_input          # 5
        json_node["auto_output"] = node.auto_output        # 6
        json_node["auto_parameter"] = node.auto_parameter  # 6b
        json_node["auto_local"] = node.auto_local          # 6c

        if node.step_size:
            json_node["step_size"] = node.step_size        # 7

        if node.children:
            json_node["container"] = [self._json_encode_node(child) for child in node.children.values()]  # 8

        if node.fmu_names_list:
            json_node["fmu"] = [f"{fmu_name}" for fmu_name in sorted(node.fmu_names_list)]          # 9

        if node.input_ports:
            json_node["input"] = [[f"{source}", f"{port.fmu_name}", f"{port.port_name}"]            # 10
                                  for port, source in node.input_ports.items()]

        if node.output_ports:
            json_node["output"] = [[f"{port.fmu_name}", f"{port.port_name}", f"{target}"]           # 11
                                   for port, target in node.output_ports.items()]

        if node.links:
            json_node["link"] = [[f"{link.from_port.fmu_name}", f"{link.from_port.port_name}",      # 12
                                  f"{link.to_port.fmu_name}", f"{link.to_port.port_name}"]
                                 for link in node.links]

        if node.start_values:
            json_node["start"] = [[f"{port.fmu_name}", f"{port.port_name}", value]                  # 13
                                  for port, value in node.start_values.items()]

        if node.drop_ports:
            json_node["drop"] = [[f"{port.fmu_name}", f"{port.port_name}"] for port in node.drop_ports]  # 14

        return json_node

    def read_ssp(self):
        logger.warning("This feature is ALPHA stage.")

        with zipfile.ZipFile(self.fmu_directory / self.filename) as zin:
            for file in zin.filelist:
                if file.filename.endswith(".fmu") or file.filename.endswith(".ssd"):
                    zin.extract(file, path=self.fmu_directory)
                    logger.debug(f"Extracted: {file.filename}")
                    self.add_transient_file(file.filename)

        self.description_pathname = self.fmu_directory / "SystemStructure.ssd"
        if self.description_pathname.is_file():
            sdd = SSDParser(step_size=self.default_step_size, auto_link=self.default_auto_link,
                            mt=self.default_mt, profiling=self.default_profiling,
                            auto_input=self.default_auto_input, auto_output=self.default_auto_output)
            self.root = sdd.parse(self.description_pathname)
            self.root.name = str(self.filename.with_suffix(".fmu"))

    def make_fmu(self, dump_json=False):
        if dump_json:
            dump_file = Path(self.input_pathname.stem + "-dump").with_suffix(".json")
            logger.info(f"Dump Json '{dump_file}'")
            self.write_json(dump_file)
        self.root.make_fmu(self.fmu_directory, debug=self.debug, description_pathname=self.description_pathname)


class SSDParser:
    def __init__(self, **kwargs):
        self.node_stack: List[AssemblyNode] = []
        self.root = None
        self.fmu_filenames: Dict[str, str] = {}  # Component name => FMU filename
        self.node_attrs = kwargs

    def parse(self, ssd_filepath: Path) -> AssemblyNode:
        logger.debug(f"Analysing {ssd_filepath}")
        with open(ssd_filepath, "rb") as file:
            parser = xml.parsers.expat.ParserCreate()
            parser.StartElementHandler = self.start_element
            parser.EndElementHandler = self.end_element
            parser.ParseFile(file)

        return self.root

    def start_element(self, tag_name, attrs):
        if tag_name == 'ssd:Connection':
            if 'startElement' in attrs:
                if 'endElement' in attrs:
                    fmu_start = self.fmu_filenames[attrs['startElement']]
                    fmu_end = self.fmu_filenames[attrs['endElement']]
                    self.node_stack[-1].add_link(fmu_start, attrs['startConnector'],
                                                 fmu_end, attrs['endConnector'])
                else:
                    fmu_start = self.fmu_filenames[attrs['startElement']]
                    self.node_stack[-1].add_output(fmu_start, attrs['startConnector'],
                                                   attrs['endConnector'])
            else:
                fmu_end = self.fmu_filenames[attrs['endElement']]
                self.node_stack[-1].add_input(attrs['startConnector'],
                                              fmu_end, attrs['endConnector'])

        elif tag_name == 'ssd:System':
            logger.info(f"SSP System: {attrs['name']}")
            filename = attrs['name'] + ".fmu"
            self.fmu_filenames[attrs['name']] = filename
            node = AssemblyNode(filename, **self.node_attrs)
            if self.node_stack:
                self.node_stack[-1].add_sub_node(node)
            else:
                self.root = node

            self.node_stack.append(node)

        elif tag_name == 'ssd:Component':
            filename = attrs['source']
            name = attrs['name']
            self.fmu_filenames[name] = filename
            self.node_stack[-1].add_fmu(filename)
            logger.debug(f"Component {name} => {filename}")

    def end_element(self, tag_name):
        if tag_name == 'ssd:System':
            self.node_stack.pop()
